! ===============================================================================================================================
! MODULE       : albedo
!
! CONTACT      : orchidee-help _at_ ipsl.jussieu.fr
!
! LICENCE      : IPSL (2006)
! This software is governed by the CeCILL licence see ORCHIDEE/ORCHIDEE_CeCILL.LIC
!
!>\BRIEF        Calculate the surface albedo using a variety of different schemes
!!
!! \n DESCRIPTION : This module includes the mechanisms for
!! 1. :: albedo_calc for computing the albedo with the old method (need to ask
!!           someone what exactly this is from)...the simplest version
!!           this does not distinguish between diffuse and direct, but can 
!!           distinguish between NIR and VIS
!! 2. :: albedo_two_stream computes the two_stream approach of Pinty et al 2006.
!!           Requires an effective leaf area index and depends on the solar angle.
!!           Separates light into NIR and VIS, and diffuse and direct illumination
!! 3. :: calculate_surface_albedo_with_snow is a way to calculate the snow albedo 
!!           based on the method of Chalita and Treut (1994)
!!           This does not distinguish between NIR and VIR light, nor diffuse 
!!           and direct, and should probably only be used with albedo_calc
!!
!! RECENT CHANGE(S): None
!!
!! REFERENCES(S)    : 
!!                    Chalita, S. and H Le Treut (1994), The albedo of temperate
!!                    and boreal forest and the Northern Hemisphere climate: a 
!!                    sensitivity experiment using the LMD GCM, 
!!                    Climate Dynamics, 10 231-240.
!!
!!                    B. Pinty, T. Lavergne, R.E. Dickinson, J-L. Widlowski, 
!!                    N. Gobron and M. M. Verstraete (2006). Simplifying the 
!!                    Interaction of Land Surfaces with Radiation for Relating 
!!                    Remote Sensing Products to Climate Models. Journal of 
!!                    Geophysical Research. Vol 111, D02116.
!!
!! SVN              :
!! $HeadURL: svn://forge.ipsl.jussieu.fr/orchidee/perso/matthew.mcgrath/TRUNK/ORCHIDEE/src_sechiba/albedo.f90 $
!! $Date: 2012-07-19 15:12:52 +0200 (Thu, 1 Nov 2012) $
!! $Revision: 947 $
!! \n
!_ ================================================================================================================================

MODULE albedo
  !
  USE ioipsl
  !
  ! modules used :
  USE constantes
  USE constantes_soil
  USE pft_parameters
  USE structures
  USE interpol_help
  USE ioipsl_para
  USE stomate_laieff

  IMPLICIT NONE

  PRIVATE :: print_debugging_albedo_info,calculate_snow_albedo,&
       twostream_solver,gammas
  ! public routines :
  PUBLIC :: albedo_calc, albedo_two_stream, calculate_surface_albedo_with_snow

  !
  ! variables used inside albedo module : declaration and initialisation
  !
                                                                        

CONTAINS

!! ==============================================================================================================================
!! SUBROUTINE   : albedo_calc
!!
!>\BRIEF        This subroutine calculates the albedo without snow.
!!
!! DESCRIPTION  : The albedo is calculated for both the visible and near-infrared 
!! domain. First the mean albedo of the bare soil is calculated. Two options exist: 
!! either the soil albedo depends on soil wetness (drysoil_frac variable), or the 
!! soil albedo is set to a mean soil albedo value.  It should be noted that this 
!! routine essential separates vegetated land into 1) 100ù plant covered and 2) 
!! bare soil parts, and calculates the albedo for them seperately.
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S): albedo 
!!
!! REFERENCE(S) : Krinner, G., Viovy, N., Noblet-Ducoudre, N. de, Ogee, J., Polcher, 
!!                J., Friedlingstein, P., Ciais, P., Sitch, S., and Prentice, I. C 
!!                (2005). A dynamic global vegetation model for studies of the 
!!                coupled atmosphere-biosphere system. Global Biogeochem. Cycles, 19, GB1015.
!!
!! FLOWCHART    : None
!! \n
!_ ================================================================================================================================

  SUBROUTINE albedo_calc (npts, drysoil_frac, veget, soilalb_dry, &
          soilalb_wet, soilalb_moy, tot_bare_soil, alb_veget, &
          alb_bare, albedo)

 !! 0. Variable and parameter declaration

    !! 0.1 Input variables
 
    INTEGER(i_std), INTENT(in)                       :: npts       !! Domain size - Number of land pixels  (unitless)
    REAL(r_std), DIMENSION(npts), INTENT(in)     :: drysoil_frac   !! Fraction of  dry soil (unitless) 
    REAL(r_std), DIMENSION(npts,nvm), INTENT(in) :: veget          !! PFT coverage fraction of a PFT (= ind*cn_ind) 
                                                                       !! (m^2 m^{-2})        
    REAL(r_std), DIMENSION(npts,2), INTENT(in)   :: soilalb_dry    !! Albedo values for the dry bare soil (unitless)
    REAL(r_std), DIMENSION(npts,2), INTENT(in)   :: soilalb_wet    !! Albedo values for the wet bare soil (unitless)
    REAL(r_std), DIMENSION(npts,2), INTENT(in)   :: soilalb_moy    !! Albedo values for the mean bare soil (unitless)
    REAL(r_std), DIMENSION (npts), INTENT(in)    :: tot_bare_soil  !! total evaporating bare soil fraction 
    
    !! 0.2 Output variables

    REAL(r_std), DIMENSION(npts,2), INTENT(out)  :: alb_veget      !! Mean vegetation albedo for visible and 
                                                                       !! near-infrared range (unitless) 
    REAL(r_std), DIMENSION(npts,2), INTENT(out)  :: alb_bare       !! Mean bare soil albedo for visible and near-infrared
    REAL(r_std),DIMENSION (npts,2), INTENT (out) :: albedo         !! Albedo for visible and near-infrared range 
                                                                       !! (unitless)   
 
    !! 0.3 Modified variables

    !! 0.4 Local variables
 
    REAL(r_std),DIMENSION (nvm,2)                    :: alb_leaf_tmp   !! Variable for albedo values for all PFTs and 
                                                                       !! spectral domains (unitless) 
    INTEGER(i_std)                                   :: ks             !! Index for visible and near-infraread range
    INTEGER(i_std)                                   :: jv             !! Index for vegetative PFTs
!_ ================================================================================================================================
    
    IF (bavard.GE.2) WRITE(numout,*) 'Entering albedo_calc'

    !! 1. Preliminary calculation
    
    ! Assign values of leaf albedo for visible and near-infrared range
    ! to local variable (constantes_veg.f90)
    alb_leaf_tmp(:,ivis) = alb_leaf_vis(:)
    alb_leaf_tmp(:,inir) = alb_leaf_nir(:)

    !! 2. Calculation and assignment of soil albedo

    DO ks = 1, 2! Loop over # of spectra

       ! If 'alb_bare_model' is set to TRUE,  the soil albedo calculation depends
       ! on soil moisture
       ! If 'alb_bare_model' is set to FALSE, the mean soil albedo is used without
       ! the dependance on soil moisture
       ! see subroutine 'conveg_soilalb'
       IF ( alb_bare_model ) THEN
          alb_bare(:,ks) = soilalb_wet(:,ks) + drysoil_frac(:) * &
               (soilalb_dry(:,ks) -  soilalb_wet(:,ks))
       ELSE
          alb_bare(:,ks) = soilalb_moy(:,ks)
       ENDIF

       ! Soil albedo is weighed by fraction of bare soil          
       albedo(:,ks) = tot_bare_soil(:) * alb_bare(:,ks)

       !! 3. Calculation of mean albedo of over the grid cell 

       ! Calculation of mean albedo of over the grid cell and
       !    mean albedo of only vegetative PFTs over the grid cell
       alb_veget(:,ks) = zero

       DO jv = 2, nvm  ! Loop over # of PFTs

          ! Mean albedo of grid cell for visible and near-infrared range
          albedo(:,ks) = albedo(:,ks) + veget(:,jv)*alb_leaf_tmp(jv,ks)

          ! Mean albedo of vegetation for visible and near-infrared range
          alb_veget(:,ks) = alb_veget(:,ks) + veget(:,jv)*alb_leaf_tmp(jv,ks)
       ENDDO ! Loop over # of PFTs

    ENDDO ! ks = 1, 2

    IF (bavard.GE.2) WRITE(numout,*) 'Leaving albedo_calc'

  END SUBROUTINE albedo_calc

!! ==============================================================================================================================\n
!! SUBROUTINE   : calculate_surface_albedo_with_snow
!!
!>\BRIEF        Calcuating snow albedo and snow cover fraction, which are then used to 
!! update the gridbox surface albedo following Chalita and Treut (1994).  This is the old
!! condveg_snow routine, and is only used for calculated the snow albedo with the old
!! albedo scheme.
!!
!! DESCRIPTION  : The snow albedo scheme presented below belongs to prognostic albedo 
!! category, i.e. the snow albedo value at a time step depends on the snow albedo value 
!! at the previous time step.
!!
!! First, the following formula (described in Chalita and Treut 1994) is used to describe 
!! the change in snow albedo with snow age on each PFT and each non-vegetative surfaces, 
!! i.e. continental ice, lakes, etc.: \n 
!! \latexonly 
!! \input{SnowAlbedo.tex}
!! \endlatexonly
!! \n
!! Where snowAge is snow age, tcstSnowa is a critical aging time (tcstSnowa=5 days)
!! snowaIni and snowaIni+snowaDec corresponds to albedos measured for aged and
!! fresh snow respectively, and their values for each PFT and each non-vegetative surfaces 
!! is precribed in in constantes_veg.f90.\n
!! In order to estimate gridbox snow albedo, snow albedo values for each PFT and 
!! each  non-vegetative surfaces with a grid box are weightedly summed up by their 
!! respective fractions.\n
!! Secondly, the snow cover fraction is computed as:
!! \latexonly 
!! \input{SnowFraction.tex}
!! \endlatexonly
!! \n
!! Where fracSnow is the fraction of snow on total vegetative or total non-vegetative
!! surfaces, snow is snow mass (kg/m^2) on total vegetated or total nobio surfaces.\n 
!! Finally, the surface albedo is then updated as the weighted sum of fracSnow, total
!! vegetated fraction, total nobio fraction, gridbox snow albedo, and previous
!! time step surface albedo.
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S): :: albedo; surface albedo. :: albedo_snow; snow
!! albedo
!!
!! REFERENCE(S) :  
!! Chalita, S. and H Le Treut (1994), The albedo of temperate and boreal forest and 
!!  the Northern Hemisphere climate: a sensitivity experiment using the LMD GCM,
!!  Climate Dynamics, 10 231-240.
!!
!! FLOWCHART    : None
!! \n
!_ ================================================================================================================================

  SUBROUTINE  calculate_surface_albedo_with_snow  (npts,  veget,  veget_max, &
           frac_nobio,  totfrac_nobio,  snow,  snow_age, &
           snow_nobio,  snow_nobio_age,  tot_bare_soil,  albedo, &
           albedo_snow,  albedo_glob)

    !! 0. Variable and parameter declaration

    !! 0.1 Input variables

    INTEGER(i_std), INTENT(in)                          :: npts        !! Domain size - Number of land pixels  (unitless)
    REAL(r_std),DIMENSION (npts,nvm), INTENT (in)   :: veget           !! PFT coverage fraction of a PFT (= ind*cn_ind) 
                                                                           !! (m^2 m^{-2})   
    REAL(r_std),DIMENSION (npts,nvm), INTENT(in)    :: veget_max
    REAL(r_std),DIMENSION (npts,nnobio), INTENT(in) :: frac_nobio      !! Fraction of non-vegetative surfaces, i.e. 
                                                                           !! continental ice, lakes, etc. (unitless)     
    REAL(r_std),DIMENSION (npts), INTENT(in)        :: totfrac_nobio   !! Total fraction of non-vegetative surfaces, i.e. 
                                                                           !! continental ice, lakes, etc. (unitless)   
    REAL(r_std),DIMENSION (npts), INTENT(in)        :: snow            !! Snow mass in vegetation (kg m^{-2})           
    REAL(r_std),DIMENSION (npts,nnobio), INTENT(in) :: snow_nobio      !! Snow mass on continental ice, lakes, etc. (kg m^{-2})      
    REAL(r_std),DIMENSION (npts), INTENT(in)        :: snow_age        !! Snow age (days)        
    REAL(r_std),DIMENSION (npts,nnobio), INTENT(in) :: snow_nobio_age  !! Snow age on continental ice, lakes, etc. (days)    
    REAL(r_std), DIMENSION (npts), INTENT(in)       :: tot_bare_soil  !! total evaporating bare soil fraction 

    !! 0.2 Output variables
    
    REAL(r_std),DIMENSION (npts,2), INTENT (inout)  :: albedo          !! Albedo (unitless ratio)          
    REAL(r_std),DIMENSION (npts), INTENT (out)      :: albedo_snow     !! Snow albedo (unitless ratio)     
    REAL(r_std),DIMENSION (npts), INTENT (out)      :: albedo_glob     !! Mean albedo (unitless ratio)     

    !! 0.3 Modified variables

    !! 0.4 Local variables
 
    INTEGER(i_std)                                      :: ji, jv, jb      !! indices (unitless)
    REAL(r_std), DIMENSION(npts)                    :: frac_snow_veg   !! Fraction of snow on vegetation (unitless ratio)
    REAL(r_std), DIMENSION(npts,nnobio)             :: frac_snow_nobio !! Fraction of snow on continental ice, lakes, etc. 
                                                                           !! (unitless ratio)
    REAL(r_std), DIMENSION(npts)                    :: snowa_veg       !! Albedo of snow covered area on vegetation
                                                                           !! (unitless ratio)
    REAL(r_std), DIMENSION(npts,nnobio)             :: snowa_nobio     !! Albedo of snow covered area on continental ice, 
                                                                           !! lakes, etc. (unitless ratio)     
    REAL(r_std), DIMENSION(npts)                    :: fraction_veg    !! Total vegetation fraction (unitless ratio)
    REAL(r_std), DIMENSION(npts)                    :: agefunc_veg     !! Age dependency of snow albedo on vegetation 
                                                                           !! (unitless)
    REAL(r_std), DIMENSION(npts,nnobio)             :: agefunc_nobio   !! Age dependency of snow albedo on ice, 
                                                                           !! lakes, .. (unitless)
    REAL(r_std)                                         :: alb_nobio       !! Albedo of continental ice, lakes, etc. 
                                                                           !!(unitless ratio)
!_ ================================================================================================================================

    IF (bavard.GE.2) WRITE(numout,*) 'Entering calculate_surface_albedo_with_snow '

    !! 1.Reset output variable (::albedo_snow ::albedo_glob) 
    
    DO ji = 1, npts
       albedo_snow(ji) = zero
       albedo_glob(ji) = zero
    ENDDO

    !! 2. Calculate snow albedos on both total vegetated and total nobio surfaces

    ! The snow albedo could be either prescribed (in condveg_init.f90) or 
    !  calculated following Chalita and Treut (1994).
    ! Check if the precribed value fixed_snow_albedo exists
    IF (ABS(fixed_snow_albedo - undef_sechiba) .GT. EPSILON(undef_sechiba)) THEN
       snowa_veg(:) = fixed_snow_albedo
       snowa_nobio(:,:) = fixed_snow_albedo
    ELSE ! calculated following Chalita and Treut (1994)

       !! 2.1 Calculate age dependence

       ! On vegetated surfaces
       DO ji = 1, npts
          agefunc_veg(ji) = EXP(-snow_age(ji)/tcst_snowa)
       ENDDO

       ! On non-vegtative surfaces
       DO jv = 1, nnobio ! Loop over # nobio types
          DO ji = 1, npts
             agefunc_nobio(ji,jv) = EXP(-snow_nobio_age(ji,jv)/tcst_snowa)
          ENDDO
       ENDDO

       !! 2.1 Calculate snow albedo 

       ! For  vegetated surfaces
       fraction_veg(:) = un - totfrac_nobio(:)
       snowa_veg(:) = zero
       IF (control%ok_dgvm) THEN
          DO ji = 1, npts
             IF ( fraction_veg(ji) .GT. min_sechiba ) THEN
                snowa_veg(ji) = snowa_veg(ji) + &
                     tot_bare_soil(ji)/fraction_veg(ji) * &
                     ( snowa_aged(1)+snowa_dec(1)*agefunc_veg(ji) )
             ENDIF
          ENDDO

          DO jv = 2, nvm
             DO ji = 1, npts
                IF ( fraction_veg(ji) .GT. min_sechiba ) THEN
                   snowa_veg(ji) = snowa_veg(ji) + &
                        veget(ji,jv)/fraction_veg(ji) * &
                        ( snowa_aged(jv)+snowa_dec(jv)*agefunc_veg(ji) )
                ENDIF
             ENDDO
          ENDDO
       ELSE
          DO jv = 1, nvm
             DO ji = 1, npts
                IF ( fraction_veg(ji) .GT. min_sechiba ) THEN
                   snowa_veg(ji) = snowa_veg(ji) + &
                        veget_max(ji,jv)/fraction_veg(ji) * &
                        ( snowa_aged(jv)+snowa_dec(jv)*agefunc_veg(ji) )
                ENDIF
             ENDDO
          ENDDO
       ENDIF ! control%ok_dgvm
       !
       ! snow albedo on other surfaces
       !
       DO jv = 1, nnobio
          DO ji = 1, npts
             snowa_nobio(ji,jv) = &
                  ( snowa_aged(1) + snowa_dec(1) * agefunc_nobio(ji,jv) ) 
          ENDDO
       ENDDO
    ENDIF ! prescribe or calculate the snow albedo?

    !! 3. Calculate snow cover fraction for both total vegetated and total 
    !! non-vegtative surfaces.\n

    ! This has been changed from the trunk version to match what is actually 
    ! in the reference...Jan 1, 2013
    frac_snow_veg(:) = MIN(MAX(snow(:),zero)/&
         (MAX(snow(:),zero)+snowcri_alb*sn_dens/100.0_r_std),un)
    DO jv = 1, nnobio
       frac_snow_nobio(:,jv) = MIN(MAX(snow_nobio(:,jv),zero)/&
            (MAX(snow_nobio(:,jv),zero)+snowcri_alb*sn_dens/100.0_r_std),un)
    ENDDO

    !! 4. Update surface albedo

    ! Update surface albedo using the weighted sum of previous time step surface
    ! albedo, total vegetated fraction, total nobio fraction, snow cover 
    ! fraction (both vegetated and non-vegetative surfaces), and snow albedo
    ! (both vegetated and non-vegetative surfaces). Although both visible and
    ! near-infrared surface albedo are presented, their calculations are the same.
    DO jb = 1, n_spectralbands

       albedo(:,jb) = ( fraction_veg(:) ) * &
            ( (un-frac_snow_veg(:)) * albedo(:,jb) + &
            ( frac_snow_veg(:)  ) * snowa_veg(:)    )
       albedo_snow(:) =  albedo_snow(:) + (fraction_veg(:)) * &
            (frac_snow_veg(:)) * snowa_veg(:)
       !
       DO jv = 1, nnobio ! Loop over # nobio surfaces
          ! 
          IF ( jv .EQ. iice ) THEN
             alb_nobio = alb_ice(jb)
          ELSE
             WRITE(numout,*) 'jv=',jv
             STOP 'DO NOT KNOW ALBEDO OF THIS SURFACE TYPE'
          ENDIF

          albedo(:,jb) = albedo(:,jb) + &
               ( frac_nobio(:,jv) ) * &
               ( (un-frac_snow_nobio(:,jv)) * alb_nobio + &
               ( frac_snow_nobio(:,jv)  ) * snowa_nobio(:,jv)   )
          albedo_snow(:) = albedo_snow(:) + &
               ( frac_nobio(:,jv) ) * ( frac_snow_nobio(:,jv) ) * &
               snowa_nobio(:,jv)
          albedo_glob(:) = albedo_glob(:) + albedo(:,jb)

       ENDDO ! jv = 1, nnobio 

    END DO ! jb = 1, n_spectralbands

    ! this accounts for the fact that the calculations are done for each
    ! spectra
    DO ji = 1, npts
       albedo_snow(ji) = (albedo_snow(ji))/REAL(n_spectralbands,r_std)
       albedo_glob(ji) = (albedo_glob(ji))/REAL(n_spectralbands,r_std)
    ENDDO

    IF (bavard.GE.2) WRITE(numout,*) 'Leaving calculate_surface_albedo_with_snow '

  END SUBROUTINE calculate_surface_albedo_with_snow
  


!! ==============================================================================================================================
!! SUBROUTINE   : albedo_two_stream
!!
!>\BRIEF        This subroutine calculates the albedo for two stream radiation transfer model.  This routine includes
!!              the effect of snow on the background albedo, through one of two methods.
!!
!! DESCRIPTION  : The albedo for two stream radiation transfer model  is calculated for both the visible and near-infrared 
!! domain. First the mean albedo of the bare soil is calculated. Two options exist: 
!! either the soil albedo depends on soil wetness (drysoil_frac variable), or the soil albedo 
!! is set to a mean soil albedo value.
!!
!!    NOTE: the main output variable, albedo, is an unweighted average of the direct and diffuse albedos
!!          for each grid point...this is done in this way right now to be consistent with the new scheme,
!!          but it doesn't have to be combined like that once we have an energy budget which can
!!          use both types of light directly
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S): ::albedo
!!
!! REFERENCE(S) : B. Pinty, T. Lavergne, R.E. Dickinson, J-L. Widlowski, N. Gobron and M. M. Verstraete (2006).
!! Simplifying the Interaction of Land Surfaces with Radiation for Relating Remote Sensing Products to Climate Models.
!! Journal of Geophysical Research. Vol 111, D02116.
!!
!! FLOWCHART    : None
!! \n
!!
!!
!!
!_ ================================================================================================================================

  SUBROUTINE albedo_two_stream(npts,  nlevels_loc, &
       drysoil_frac,  lai,  &
       veget_max,  sinang,   soilalb_dry,  soilalb_wet, &
       frac_nobio,  soilalb_moy,  alb_bare,  albedo, &
       snow,  snow_age,  snow_nobio,  snow_nobio_age, &
       albedo_snow,  albedo_glob, circ_class_biomass, &
       circ_class_n,  lai_split, z0_veg, Isotrop_Abs_Tot_p,&
       Isotrop_Tran_Tot_p, laieff_fit, albedo_pft, frac_snow_nobio,&
       frac_snow_veg)

 !! 0. Variable and parameter declaration

    !! 0.1 Input variables
 
   INTEGER(i_std), INTENT(in)                                :: npts        !! Domain size - Number of land pixels  (unitless)       
   INTEGER(i_std), INTENT(in)                                :: nlevels_loc     !! The number of vertical levels in the canopy (unitless)
   REAL(r_std), DIMENSION(:), INTENT(in)              :: drysoil_frac    !! Fraction of  dry soil (unitless)
   REAL(r_std),DIMENSION (:,:), INTENT (in)         :: lai             !! Leaf area index (m^2 m^{-2})

   REAL(r_std), DIMENSION(:), INTENT(in)              :: sinang          !! 
   REAL(r_std), DIMENSION(:,:), INTENT(in)          :: veget_max       !! PFT coverage fraction of a PFT (= ind*cn_ind) 
                                                                                !! (m^2 m^{-2})
   REAL(r_std), DIMENSION(:,:), INTENT(in)            :: soilalb_dry     !! Albedo values for the dry bare soil (unitless)
   REAL(r_std), DIMENSION(:,:), INTENT(in)            :: soilalb_wet     !! Albedo values for the wet bare soil (unitless)
   REAL(r_std), DIMENSION(:,:), INTENT(in)            :: soilalb_moy     !! Albedo values for the mean bare soil (unitless)
   REAL(r_std),DIMENSION (:), INTENT(in)              :: snow            !! Snow mass in vegetation (kg m^{-2})           
   REAL(r_std),DIMENSION (:,:), INTENT(in)       :: snow_nobio      !! Snow mass on continental ice, lakes, etc. (kg m^{-2})      
   REAL(r_std),DIMENSION (:), INTENT(in)              :: snow_age        !! Snow age (days)        
   REAL(r_std),DIMENSION (:,:), INTENT(in)       :: snow_nobio_age  !! Snow age on continental ice, lakes, etc. (days)    
   REAL(r_std),DIMENSION (:,:), INTENT(in)       :: frac_nobio      !! Fraction of non-vegetative surfaces, i.e. 
                                                                                !! continental ice, lakes, etc. (unitless)     
   REAL(r_std), DIMENSION(:,:,:,:,:), &
                                               INTENT(IN)    :: circ_class_biomass !! Stem diameter 
                                                                                !! @tex $(m)$ @endtex
   REAL(r_std), DIMENSION(:,:,:), INTENT(IN)    :: circ_class_n    !! Number of trees within each circumference
   REAL(r_std),DIMENSION(:),INTENT(IN)                 :: lai_split
   REAL(r_std), DIMENSION(:), INTENT(in)              :: z0_veg          !! Roughness height of vegetated part (m)
   TYPE(laieff_type),DIMENSION (:,:,:),INTENT(in)    :: laieff_fit      !! Fitted parameters for the effective LAI

   
   !! 0.2 Output variables
   
   REAL(r_std), DIMENSION(:,:), INTENT(out)           :: alb_bare        !! Mean bare soil albedo for visible and near-infrared
   REAL(r_std), DIMENSION (:,:), INTENT (out)         :: albedo          !! Albedo (two stream radiation transfer model)
                                                                                !! for visible and near-infrared range 
                                                                                !! (unitless)
   REAL(r_std),DIMENSION (:), INTENT (out)            :: albedo_snow     !! Snow albedo (unitless ratio)     
   REAL(r_std),DIMENSION (:), INTENT (out)            :: albedo_glob     !! Mean albedo (unitless ratio)
   REAL(r_std),DIMENSION (:,:,:), INTENT (out)        :: Isotrop_Abs_Tot_p !! Absorbed radiation per layer for photosynthesis
   REAL(r_std),DIMENSION (:,:,:), INTENT (out)        :: Isotrop_Tran_Tot_p !! Transmitted radiation per layer for photosynthesis
   REAL(r_std), DIMENSION (:,:,:), INTENT (out)       :: albedo_pft      !! Albedo (two stream radiation transfer model)
                                                                         !! for visible and near-infrared range 
                                                                         !! for each PFT (unitless)
   REAL(r_std), DIMENSION(npts,nnobio), INTENT(out)   :: frac_snow_nobio !! Fraction of snow on continental ice, lakes, etc. 
                                                                         !! (unitless ratio)
   REAL(r_std), DIMENSION(npts,nvm), INTENT(out)      :: frac_snow_veg   !! the fraction of ground covered with snow, between 0 and 1



   !! 0.3 Modified variables
   
   !! 0.4 Local variables

   REAL(r_std), DIMENSION(nlevels_loc,npts,nvm)          :: laieff_collim  !! Leaf Area Index Effective for direct light
   REAL(r_std), DIMENSION(nlevels_loc,npts,nvm)          :: laieff_isotrop  !! Leaf Area Index Effective converts
                                                                                !! 3D lai into 1D lai for two stream
                                                                                !! radiation transfer model...this is for 
                                                                                !! isotropic light and only calculated once per day
                                                                                !! @tex $(m^{2} m^{-2})$ @endtex      
   REAL(r_std),DIMENSION (nlevels_loc)         :: laieff_isotrop_pft, laieff_collim_pft   !! 
   REAL(r_std)                           :: cosine_sun_angle   !! the cosine of the solar zenith angle
   INTEGER(i_std)                        :: ks                 !! Index for visible and near-infraread range
   INTEGER(i_std)                        :: ivm                !! Index for vegetative PFTs
   INTEGER(i_std)                        :: ipts                !! Index for spatial domain
   INTEGER(i_std)                        :: ilevel             !! Index for canopy levels
   REAL(r_std)                           :: leaf_reflectance
   REAL(r_std)                           :: leaf_transmittance
   REAL(r_std)                           :: br_base_temp,br_base_temp_collim,br_base_temp_isotrop
   REAL(r_std)                           :: leaf_psd_temp
   REAL(r_std)                           :: leaf_single_scattering_albedo
   REAL(r_std),DIMENSION(nlevels_loc)          :: Collim_Alb_Tot   !! Collimated (direct) total albedo for a given level
   REAL(r_std),DIMENSION(nlevels_loc)          :: Collim_Tran_Uncoll  !! collimated total transmission of uncollided light for a given level 
   REAL(r_std),DIMENSION(nlevels_loc)          :: Collim_Tran_Coll  !! collimated total transmission for a given level
   REAL(r_std),DIMENSION(nlevels_loc)          :: Collim_Abs_Tot   !! collimated total absorption for a given level
   REAL(r_std),DIMENSION(nlevels_loc)          :: Isotrop_Alb_Tot  !! isotropic (diffuse) total albedo for a given level
   REAL(r_std),DIMENSION(nlevels_loc)          :: Isotrop_Tran_Uncoll !! isotropic total transmission of uncollided light for a given level
   REAL(r_std),DIMENSION(nlevels_loc)          :: Isotrop_Tran_Coll !! isotropic total transmission for a given level
   REAL(r_std),DIMENSION(nlevels_loc)          :: Isotrop_Abs_Tot  !! isotropic total absporbtion for a given level
   INTEGER :: jv, inno
   REAL(r_std) :: alb_nobio

   REAL(r_std):: laieff_collim_1, laieff_isotrop_1, Collim_Alb_Tot_1,Collim_Tran_Tot_1,Collim_Abs_Tot_1,&
                 Isotrop_Alb_Tot_1,Isotrop_Tran_Tot_1,Isotrop_Abs_Tot_1,&
                 Collim_Tran_Uncollided_1, Isotrop_Tran_Uncollided_1   !! Values for the one level solution

   REAL(r_std):: converged_albedo
   REAL(r_std):: isotrop_angle
   LOGICAL :: lconverged
   LOGICAL :: lprint           !! a flag for printing some debug statements
   REAL(r_std), DIMENSION(npts,nvm,n_spectralbands) :: snowa_veg  !! snow albedo due to vegetated surfaces, between 0 and 1
   REAL(r_std), DIMENSION(npts,nnobio,n_spectralbands) :: snowa_nobio !! Albedo of snow covered area on continental ice, 
                                                                          !! lakes, etc. (unitless ratio)  
   
!_ ================================================================================================================================


   IF (bavard.GE.2) WRITE(numout,*) 'Entering albedo_two_stream'

   !! 1. We will now calculate the background reflectance to be used in the model.
   !! The parameters read in from the input file do not include the effect of snow,
   !! so we need to figure out the snow's contribution for each grid space
   
   !initialize some output variables
   DO ipts = 1, npts
      albedo_snow(ipts) = zero
   ENDDO
   Isotrop_Abs_Tot_p(:,:,:)=zero
   Isotrop_Tran_Tot_p(:,:,:)=un
   albedo_pft(:,:,:)=zero

   ! We need to calculate the LAI effective for this solar angle. The other LAI
   ! effective that we calculated was computed at an angle of 60.0 degrees for
   ! the isotropic contribution. Notice that, in many situations, the results 
   ! will be exactly the same.  One can show that Pgap is proportional to 
   ! 1/cos(theta) if the level bottom is below all canopy elements, which
   ! means that the cos(theta) in the calculation of the LAIeff will be
   ! cancelled out

   isotrop_angle=COS(60.0/180.0*pi)
   DO ipts=1,npts
      DO ivm=1,nvm
         DO ilevel=1,nlevels_loc
            laieff_isotrop(ilevel,ipts,ivm)=calculate_laieff_fit(isotrop_angle,laieff_fit(ipts,ivm,ilevel))
            laieff_collim(ilevel,ipts,ivm)=calculate_laieff_fit(sinang(ipts),laieff_fit(ipts,ivm,ilevel))
            ! Because we are fitting these values, it's possible that some will 
            ! drop below zero.  This causes a serious problem here,
            ! so let's make sure this doesn't happen.  Unfortunately, by
            ! doing this, we lose a test of things going wrong (an unfitted
            ! negative LAIeff can be the sign of another problem), but we
            ! don't really have a choice.
            IF(laieff_isotrop(ilevel,ipts,ivm) .LT. min_stomate) &
                 laieff_isotrop(ilevel,ipts,ivm)=zero
            IF(laieff_collim(ilevel,ipts,ivm) .LT. min_stomate) &
                 laieff_collim(ilevel,ipts,ivm)=zero
         ENDDO
      ENDDO
   ENDDO

   IF(ld_albedo)THEN
      DO ivm=1,nvm
         IF(ivm == test_pft)THEN
            DO ipts = 1, npts
               IF(ipts == test_grid)THEN
                  WRITE(numout,*) 'Laieff_collim and laieff_isotrop in albedo'
                  WRITE(numout,*) 'sinang(ipts): ',ACOS(sinang(ipts))/pi*180.0_r_std
                  DO ilevel=nlevels_loc,1,-1
                     WRITE(numout,*) ilevel,laieff_collim(ilevel,ipts,ivm),&
                          laieff_isotrop(ilevel,ipts,ivm)
                  END DO
                  DO ilevel=nlevels_loc,1,-1
                     WRITE(numout,*) 'Fitting parameters: ',laieff_fit(ipts,ivm,ilevel)%a,laieff_fit(ipts,ivm,ilevel)%b,&
                          laieff_fit(ipts,ivm,ilevel)%c,laieff_fit(ipts,ivm,ilevel)%d,laieff_fit(ipts,ivm,ilevel)%e
                  ENDDO
               ENDIF
            ENDDO
         ENDIF
      ENDDO
   ENDIF

   CALL calculate_snow_albedo(npts,  sinang,  snow,  snow_nobio, &
        snow_age, snow_nobio_age,  frac_nobio,  albedo_snow, &
        snowa_veg,  frac_snow_veg,  snowa_nobio,  frac_snow_nobio, &
        veget_max, z0_veg)

!!$   !+++++ CHECK +++++++
!!$   DO ipts=1,npts
!!$      DO ivm=1,nvm
!!$         IF(frac_snow_veg(ipts,ivm) .LT. 0.0)THEN
!!$            WRITE(numout,*) 'SNOWFRAC ipts ',ipts,frac_snow_veg(ipts,ivm)
!!$            CALL ipslerr_p (3,'albedo', 'snow frac error','','')
!!$         ENDIF
!!$      ENDDO
!!$      DO inno=1,nnobio
!!$         IF(frac_snow_nobio(ipts,inno) .LT. 0.0)THEN
!!$            WRITE(numout,*) 'SNOWFRAC ipts ',ipts,frac_snow_nobio(ipts,inno)
!!$            CALL ipslerr_p (3,'albedo', 'nobio snow frac error','','')
!!$         ENDIF
!!$      ENDDO
!!$   ENDDO
!!$
!!$   !++++++++++++++++++

   ! initialize the albedo in every square for both spectra by calculating the 
   ! bare soil albedo

   DO ks = 1, n_spectralbands 
      DO ipts = 1, npts
         ! If 'alb_bare_model' is set to TRUE,  the soil albedo calculation 
         ! depends on soil moisture
         ! If 'alb_bare_model' is set to FALSE, the mean soil albedo is used 
         ! without the dependance on soil moisture
         ! see subroutine 'conveg_soilalb'
         IF ( alb_bare_model ) THEN
            alb_bare(ipts,ks) = soilalb_wet(ipts,ks) + drysoil_frac(ipts) * &
                 (soilalb_dry(ipts,ks) -  soilalb_wet(ipts,ks))
         ELSE
            alb_bare(ipts,ks) = soilalb_moy(ipts,ks)
         ENDIF

         ! we need to take into account any snow that has fallen on the bare soil
         ! since there may be some bare soil, initialize the albedo with this fraction
         albedo(ipts,ks) = veget_max(ipts,1) * (alb_bare(ipts,ks)*&
              (un-frac_snow_veg(ipts,1))+frac_snow_veg(ipts,1)*snowa_veg(ipts,1,ks))
      ENDDO ! ipts = 1, npts
   ENDDO ! ks = 1, n_spectralbands 

   !! 3. Calculation of albedo of the whole grid cell, taking into account all 
   !!    PFTs (except bare soil) and spectral bands

   DO ipts = 1, npts ! loop over the grid squares

      ! it appears that sinang is the sin of the angle from the horizon
      ! this is equal to the cos of the zenith angle, which is what we need
      cosine_sun_angle = sinang(ipts)

      ! For the coupled model, the angles can be negative.
      ! Offline, they are always between zero and 90 degrees.
      IF(cosine_sun_angle .LE. min_sechiba)THEN
         ! it's night, so no sunlight...don't need to calculate albedo.  Set it equal to
         ! zero, because the snow albedo has already been calculated and it looks funny on
         ! the coupled run maps otherwise.
         albedo(ipts,:)=zero
         CYCLE
      ENDIF


      ! are there any non-bio PFTs here that we need to take into account?
      DO ks = 1, n_spectralbands ! Loop over # of spectra
         DO jv = 1, nnobio 
            ! now update the albedo
            IF ( jv .EQ. iice ) THEN
               alb_nobio = alb_ice(ks)
            ELSE
               WRITE(numout,*) 'jv=',jv
               STOP 'DO NOT KNOW ALBEDO OF THIS SURFACE TYPE'
            ENDIF

            ! this takes into account both snow covered and non-snow covered non
            ! bio regios in all grid squares
            albedo(ipts,ks) = albedo(ipts,ks) + &
                 ( frac_nobio(ipts,jv) ) * &
                 ( (un-frac_snow_nobio(ipts,jv)) * alb_nobio + &
                 ( frac_snow_nobio(ipts,jv)  ) * snowa_nobio(ipts,jv,ks)   )
            albedo_snow(ipts) = albedo_snow(ipts) + &
                 ( frac_nobio(ipts,jv) ) * ( frac_snow_nobio(ipts,jv) ) * &
                 snowa_nobio(ipts,jv,ks)/REAL(n_spectralbands,r_std) 

         ENDDO ! jv = 1, nnobio 
      ENDDO ! ks = 1, n_spectralbands 

      DO ivm = 2, nvm  ! Loop over # of PFTs     

         ! for this grid square and this vegetation type, we have a set LAI 
         ! effective
         laieff_collim_pft(1:nlevels_loc) = laieff_collim(1:nlevels_loc,ipts,ivm)
         laieff_isotrop_pft(1:nlevels_loc) = laieff_isotrop(1:nlevels_loc,ipts,ivm)

         IF(veget_max(ipts,ivm) == zero)THEN
            ! this vegetation type is not present, so no reason to do the 
            ! calculation
            CYCLE
         ENDIF

         DO ks = 1, n_spectralbands ! Loop over # of spectra

            ! set single scattering albedo and preferred scattering direction
            leaf_single_scattering_albedo = leaf_ssa(ivm,ks)
            leaf_psd_temp = leaf_psd(ivm,ks)                 

            ! calculate the rleaf and tleaf from wl=rl+tl and dl=rl/tl

            leaf_transmittance = leaf_single_scattering_albedo/ & 
                 (leaf_psd_temp+1)
            leaf_reflectance = leaf_psd_temp * &
                 leaf_transmittance

            ! We need to take into account the effect of snow on the background 
            ! reflectance here

            ! from the snow above I can calculate the true background reflectance
            ! for this PFT
            br_base_temp= (un-frac_snow_veg(ipts,ivm)) * bgd_reflectance(ivm,ks)&
                 + ( frac_snow_veg(ipts,ivm)  ) * snowa_veg(ipts,ivm,ks)

            ! At this step, we assume that there is no difference between the 
            ! direct and the diffuse background reflectance, which is true for 
            ! the background but not the canopy albedo
            br_base_temp_collim=br_base_temp
            br_base_temp_isotrop=br_base_temp

            ! This prints out some debugging information
            IF(ld_albedo .AND. ivm == test_pft .AND. ipts == test_grid)THEN
               DO ilevel=nlevels_loc,1,-1
                  laieff_collim_1=laieff_collim_pft(ilevel)
                  laieff_isotrop_1=laieff_isotrop_pft(ilevel)
                  call twostream_solver(leaf_reflectance,leaf_transmittance,&
                       br_base_temp_collim,br_base_temp_isotrop,&
                       cosine_sun_angle,Collim_Alb_Tot_1,Collim_Tran_Tot_1,&
                       Collim_Abs_Tot_1,Isotrop_Alb_Tot_1,Isotrop_Tran_Tot_1,&
                       Isotrop_Abs_Tot_1,laieff_collim_1,  laieff_isotrop_1, &
                       Collim_Tran_Uncollided_1, Isotrop_Tran_Uncollided_1)
                  
                  
                  CALL print_debugging_albedo_info(ilevel,cosine_sun_angle,&
                       leaf_reflectance,leaf_transmittance,&
                       laieff_collim_1,laieff_isotrop_1,&
                       br_base_temp_collim,br_base_temp_isotrop,Collim_Alb_Tot_1,&
                       Collim_Tran_Tot_1,Collim_Abs_Tot_1,&
                       Isotrop_Alb_Tot_1,Isotrop_Tran_Tot_1,Isotrop_Abs_Tot_1)
                  
               ENDDO ! loop over levels
            ENDIF ! debugging

            ! Calculate the one level approach to use as a template. Notice that
            ! this is not actually needed for the multilevel approach, since there
            ! are no parameters that we can tune to get the multilevel result
            ! closer to the one level result.

            ! First, we need to recombine all the levels into a one level LAI
            laieff_collim_1=SUM(laieff_collim_pft(1:nlevels_loc))
            laieff_isotrop_1=SUM(laieff_isotrop_pft(1:nlevels_loc))
            call twostream_solver(leaf_reflectance, leaf_transmittance, &
                 br_base_temp_collim, br_base_temp_isotrop, &
                 cosine_sun_angle, Collim_Alb_Tot_1, Collim_Tran_Tot_1, &
                 Collim_Abs_Tot_1, &
                 Isotrop_Alb_Tot_1, Isotrop_Tran_Tot_1, Isotrop_Abs_Tot_1, &
                 laieff_collim_1, &
                 laieff_isotrop_1,  Collim_Tran_Uncollided_1,  &
                 Isotrop_Tran_Uncollided_1)

             ! Print out more debugging information
             IF(ld_albedo .AND. ipts == test_grid .AND. ivm == test_pft) &
                  CALL print_debugging_albedo_info(1, cosine_sun_angle, &
                  leaf_reflectance, leaf_transmittance, laieff_collim_1, &
                  laieff_isotrop_1, br_base_temp_collim, br_base_temp_isotrop, &
                  Collim_Alb_Tot_1, Collim_Tran_Tot_1, Collim_Abs_Tot_1, &
                  Isotrop_Alb_Tot_1, Isotrop_Tran_Tot_1, Isotrop_Abs_Tot_1)

            ! If we only have one canopy level, there is no reason to try to fit
            ! the nlevel case, since we just calculated the answer above
            IF(nlevels_loc == 1)THEN
               ! notice that this albedo assumes that diffuse and direct albedo 
               ! are of equal importance
               converged_albedo=(Collim_Alb_Tot_1+Isotrop_Alb_Tot_1)/2.0_r_std
               albedo(ipts,ks) = albedo(ipts,ks) + veget_max(ipts,ivm)*converged_albedo
               albedo_pft(ipts,ivm,ks)=veget_max(ipts,ivm)*converged_albedo
               Isotrop_Abs_Tot_p(ipts,ivm,:) = Isotrop_Abs_Tot_1
               Isotrop_Tran_Tot_p(ipts,ivm,:) = Isotrop_Tran_Tot_1

               CYCLE
            ENDIF


            ! now solve the multilevel scheme
            IF(ld_albedo .AND. test_pft == ivm .AND. test_grid == ipts)THEN
               lprint=.TRUE.
            ELSE
               lprint=.FALSE.
            ENDIF

            CALL multilevel_albedo(nlevels_loc, cosine_sun_angle, leaf_single_scattering_albedo, &
                 leaf_psd_temp, br_base_temp_collim, br_base_temp_isotrop, &
                 laieff_collim_pft, laieff_isotrop_pft, lconverged, &
                 Collim_Alb_Tot, Collim_Tran_Coll, Collim_Abs_Tot, Isotrop_Alb_Tot, &
                 Isotrop_Tran_Coll, Isotrop_Abs_Tot, Collim_Tran_Uncoll, Isotrop_Tran_Uncoll, lprint)
            
            ! This is a lot of information printed out for the paper
            ! on the multilevel albedo scheme.  Probably not useful
            ! for anyone else, but I'm keeping it in until the paper is published.
            ! This is why I'm protecting it with an IF statement.  I do not use
            ! ld_albedo since it really is too much information for normal usage.
            IF(.FALSE.)THEN

               WRITE(6,'(A,2F14.10)') 'Background reflectance ',&
                    br_base_temp_collim,br_base_temp_isotrop
               WRITE(6,'(A,F14.10)') 'Single scattering albedo (wl): ',&
                    leaf_single_scattering_albedo
               WRITE(6,'(A,F14.10)') 'Preferred scattering direction (dl): ',&
                    leaf_psd_temp

               WRITE(6,*) 'Collimated light ',ACOS(cosine_sun_angle)/pi*180.0
               WRITE(6,'(A,F4.1,A,4(F10.6,A))') '        1  &  ',laieff_collim_1,&
                    ' & ',Collim_Alb_Tot_1,' & ',&
                    Collim_Tran_Tot_1-Collim_Tran_Uncollided_1,' & ',&
                    Collim_Tran_Uncollided_1,' & ',&
                    Collim_Alb_Tot_1+Collim_Tran_Tot_1,' \\ '
               WRITE(6,'(A,F4.1,A,4(F10.6,A))') '        2  &  ',laieff_collim_1,&
                    ' & ',Collim_Alb_Tot(nlevels_loc),' & ',&
                    Collim_Tran_Coll(1),' & ',Collim_Tran_Uncoll(1),' & ',&
                    Collim_Alb_Tot(nlevels_loc)+Collim_Tran_Coll(1)+&
                    Collim_Tran_Uncoll(1),' \\ '
               WRITE(6,'(A,F4.1,A,4(F10.6,A))') '     Diff  &  ',laieff_collim_1,&
                    ' & ',ABS(Collim_Alb_Tot_1-Collim_Alb_Tot(nlevels_loc)),' & ',&
                    ABS((Collim_Tran_Tot_1-Collim_Tran_Uncollided_1)-&
                    Collim_Tran_Coll(1)),' & ',&
                    ABS(Collim_Tran_Uncollided_1-Collim_Tran_UnColl(1)),' & ',&
                    ABS((Collim_Alb_Tot_1+Collim_Tran_Tot_1)-&
                    (Collim_Alb_Tot(nlevels_loc)+Collim_Tran_Coll(1)+&
                    Collim_Tran_Uncoll(1))),' \\ '

               WRITE(6,*) 'Isotropic light '
               WRITE(6,'(A,F4.1,A,4(F10.6,A))') '        1  &  ',&
                    laieff_isotrop_1,' & ',Isotrop_Alb_Tot_1,' & ',&
                    Isotrop_Tran_Tot_1-Isotrop_Tran_Uncollided_1,' & ',&
                    Isotrop_Tran_Uncollided_1,' & ',&
                    Isotrop_Alb_Tot_1+Isotrop_Tran_Tot_1,' \\ '
               WRITE(6,'(A,F4.1,A,4(F10.6,A))') '        2  &  ',&
                    laieff_isotrop_1,' & ',Isotrop_Alb_Tot(nlevels_loc),' & ',&
                    Isotrop_Tran_Coll(1),' & ',Isotrop_Tran_Uncoll(1),' & ',&
                    Isotrop_Alb_Tot(nlevels_loc)+Isotrop_Tran_Coll(1)+&
                    Isotrop_Tran_Uncoll(1),' \\ '
               WRITE(6,'(A,F4.1,A,4(F10.6,A))') '     Diff  &  ',&
                    laieff_isotrop_1,&
                    ' & ',ABS(Isotrop_Alb_Tot_1-Isotrop_Alb_Tot(nlevels_loc)),' & ',&
                    ABS((Isotrop_Tran_Tot_1-Isotrop_Tran_Uncollided_1)-&
                    Isotrop_Tran_Coll(1)),' & ',&
                    ABS(Isotrop_Tran_Uncollided_1-Isotrop_Tran_UnColl(1)),' & ',&
                    ABS((Isotrop_Alb_Tot_1+Isotrop_Tran_Tot_1)-&
                    (Isotrop_Alb_Tot(nlevels_loc)+Isotrop_Tran_Coll(1)+&
                    Isotrop_Tran_Uncoll(1))),' \\ '

               WRITE(6,1010) 'COLLIMTRAN ',laieff_collim_1, &
                    laieff_collim_pft(nlevels_loc),1,&
                    leaf_single_scattering_albedo,br_base_temp_collim,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    Collim_Tran_Tot_1,(Collim_Tran_Coll(1)+Collim_Tran_Uncoll(1))
               WRITE(6,1010) 'COLLIMALB ',laieff_collim_1, &
                    laieff_collim_pft(nlevels_loc),2,&
                    leaf_single_scattering_albedo,br_base_temp_collim,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    Collim_Alb_Tot_1,Collim_Alb_Tot(nlevels_loc)
               WRITE(6,1010) 'COLLIMABSCAN ',laieff_collim_1, &
                    laieff_collim_pft(nlevels_loc),3,&
                    leaf_single_scattering_albedo,br_base_temp_collim,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    Collim_Abs_Tot_1,SUM(Collim_Abs_Tot(:))
               WRITE(6,1010) 'COLLIMABSSOIL ',laieff_collim_1, &
                    laieff_collim_pft(nlevels_loc),4,&
                    leaf_single_scattering_albedo,br_base_temp_collim,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    un-Collim_Alb_Tot_1-Collim_Abs_Tot_1,un-&
                    Collim_Alb_Tot(nlevels_loc)-SUM(Collim_Abs_Tot(:))

               WRITE(6,1010) 'ISOTROPTRAN ',laieff_isotrop_1, &
                    laieff_isotrop_pft(nlevels_loc),1,&
                    leaf_single_scattering_albedo,br_base_temp_isotrop,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    Isotrop_Tran_Tot_1,(Isotrop_Tran_Coll(1)+Isotrop_Tran_Uncoll(1))
               WRITE(6,1010) 'ISOTROPALB ',laieff_isotrop_1, &
                    laieff_isotrop_pft(nlevels_loc),2,&
                    leaf_single_scattering_albedo,br_base_temp_isotrop,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    Isotrop_Alb_Tot_1,Isotrop_Alb_Tot(nlevels_loc)
               WRITE(6,1010) 'ISOTROPABSCAN ',laieff_isotrop_1, &
                    laieff_isotrop_pft(nlevels_loc),3,&
                    leaf_single_scattering_albedo,br_base_temp_isotrop,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    Isotrop_Abs_Tot_1,SUM(Isotrop_Abs_Tot(:))
               WRITE(6,1010) 'ISOTROPABSSOIL ',laieff_isotrop_1, &
                    laieff_isotrop_pft(nlevels_loc),4,&
                    leaf_single_scattering_albedo,br_base_temp_isotrop,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    un-Isotrop_Alb_Tot_1-Isotrop_Abs_Tot_1,un-&
                    Isotrop_Alb_Tot(nlevels_loc)-SUM(Isotrop_Abs_Tot(:))

1010           FORMAT(A,2(F12.6,2X),I4,2X,6(F14.8,2X))

            ENDIF ! debugging

            ! For our total albedo, we are just taking a simple mean of the diffuse
            ! and the direct light, which means we lose some information. This 
            ! might be changed in the future. We also need to weight this by the 
            ! percentage of nonbio land, but this appears to be included in veget_max

            converged_albedo =(Collim_Alb_Tot(nlevels_loc)+Isotrop_Alb_Tot(nlevels_loc))/2.0_r_std
            albedo(ipts,ks) = albedo(ipts,ks) + veget_max(ipts,ivm)*converged_albedo
            albedo_pft(ipts,ivm,ks)=veget_max(ipts,ivm)*converged_albedo

            ! Save the absorbed radiation for photosynthesis, we need only the
            ! visible range and use the isotropic part, this can be change later
            ! to distinguish between sunlit and shaded leaves.
            ! Be careful here
            IF (ks == 1) THEN
               ! Test if Collim_Abs_Tot has negative values
               ! If so, set it to min_sechiba  
               DO  ilevel=1,nlevels_loc
                  IF (Isotrop_Abs_Tot(ilevel) .LT. zero) THEN
                     Isotrop_Abs_Tot(ilevel) =  min_sechiba
                  ENDIF
               ENDDO
               !
               Isotrop_Abs_Tot_p(ipts,ivm,:) = Isotrop_Abs_Tot(:)
               Isotrop_Tran_Tot_p(ipts,ivm,:) = Isotrop_Tran_Coll(:) + Isotrop_Tran_Uncoll(:)

               ! Notice that this light is actually cumulative, not per
               ! level!  This was needed for debugging purposes and running
               ! tests.  However, the photosynthesis routines need the light
               ! transmitted per level, i.e. if there is zero LAI
               ! in a level, the transmitted light will be one.
               DO ilevel=1,nlevels_loc-1
                  IF(Isotrop_Tran_Tot_p(ipts,ivm,ilevel+1) .GT. alb_threshold)THEN
                     Isotrop_Tran_Tot_p(ipts,ivm,ilevel)=&
                          Isotrop_Tran_Tot_p(ipts,ivm,ilevel)/Isotrop_Tran_Tot_p(ipts,ivm,ilevel+1)
                  ELSE
                     ! Here, we really don't know anything about how much light is transmitted
                     ! in this layer, but there is no light reaching it from above so we
                     ! can safely assume no photosynthesis takes place.  This is equivalent
                     ! to assuming it has no LAI, which means the transmission will be unity.
                     Isotrop_Tran_Tot_p(ipts,ivm,ilevel)=un
                  ENDIF
               ENDDO

               ! Debugging information
               IF (ld_albedo .AND. ivm == test_pft .AND. ipts == test_grid) THEN
                  WRITE(numout,*) 'Absorbed light in albedo.f90: ', ipts, ivm
                  DO ilevel=nlevels_loc,1,-1
                     WRITE(numout,'(I5,10F20.10)') ilevel, Isotrop_Abs_Tot_p(ipts,ivm,ilevel),&
                          Isotrop_Tran_Tot_p(ipts,ivm,ilevel),&
                          laieff_isotrop(ilevel,ipts,ivm),laieff_collim(ilevel,ipts,ivm)
                  ENDDO
               ENDIF

            ENDIF ! IF ks==1

         ENDDO ! ks = 1, n_spectralbands 

      ENDDO ! ivm = 2, nvm  

   ENDDO ! ipts = 1, npts 

   ! now we need to average the albedo over all our spectra, so we can pass it to
   ! methods which do not distinguish between different spectral bands
   albedo_glob(:) = zero
   DO ks = 1, n_spectralbands ! Loop over # of spectra
      albedo_glob(:) = albedo_glob(:) + albedo(:,ks)
   ENDDO
   albedo_glob(:)=albedo_glob(:)/REAL(n_spectralbands,r_std)

   ! WRITE(*,*) '281013 in albedo_two_stream, Isotrop_Abs_Tot_p is: ', Isotrop_Abs_Tot_p

   IF (bavard.GE.2) WRITE(numout,*) 'Leaving albedo_two_stream'

 END SUBROUTINE albedo_two_stream

!! ==============================================================================================================================
!! SUBROUTINE   : twostream_solver
!!
!>\BRIEF        : Computes the two-stream albedo solution for a level with given single scatterer properties
!!                and a defined background albedo
!!
!! DESCRIPTION  : This solution is the two-stream solution of Pinty et al (2006) for a vegetation level above
!!                an isotropically reflecting background.  It breaks down the problem into three parts, solving
!!                each part for the case of diffuse (isotropic) and direct (collimated) light.  The three parts are,
!!                1) The term due to light that does not interact at all with the canopy (Black Canopy)
!!                2) Light that does not interact at all with the background (Black Background)
!!                3) Light that bounces between the background and the canopy
!!                This routine (and the routines it uses) were received directly from Bernard Pinty, and only some
!!                minor modifcations were made, in addition to more documentation
!! 
!! 
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S): Collim_Alb_Tot,Collim_Tran_Tot,
!!      Collim_Abs_Tot,Isotrop_Alb_Tot,Isotrop_Tran_Tot,Isotrop_Abs_To
!!
!! REFERENCE(S) :  B. Pinty, T. Lavergne, R.E. Dickinson, J-L. Widlowski, N. Gobron and M. M. Verstraete (2006).
!! Simplifying the Interaction of Land Surfaces with Radiation for Relating Remote Sensing Products to Climate Models.
!! Journal of Geophysical Research. Vol 111, D02116.
!!
!! FLOWCHART    : None
!! \n
!_ ================================================================================================================================


 SUBROUTINE twostream_solver(leaf_reflectance, leaf_transmittance, background_reflectance_collim, background_reflectance_isotrop, &
          cosine_sun_angle, Collim_Alb_Tot, Collim_Tran_Tot, Collim_Abs_Tot, &
          Isotrop_Alb_Tot, Isotrop_Tran_Tot, Isotrop_Abs_Tot, laieff_collim, &
           laieff_isotrop,  Collim_Tran_Uncollided,  Isotrop_Tran_Uncollided)


   !! 0. Variables and parameter declaration
   !! 0.1 Input variables
   REAL(r_std), INTENT(IN)    :: leaf_reflectance            !! effective leaf reflectance, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(IN)    :: leaf_transmittance          !! effective leaf transmittance, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(IN)    :: background_reflectance_collim      !! background reflectance for direct radiation,
                                                             !! between 0 and 1
   REAL(r_std), INTENT(IN)    :: background_reflectance_isotrop      !! background reflectance for diffuse radiation,
                                                             !! between 0 and 1
   REAL(r_std), INTENT(IN)    :: cosine_sun_angle            !! cosine of the solar zenith angle, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(IN)    :: laieff_collim               !! Effective Leaf Area Index, computed at current sun angle
   REAL(r_std), INTENT(IN)    :: laieff_isotrop              !! Effective Leaf Area Index, computed at 60 degrees
                                                             !! @tex $(m^{2} m^{-2})$ @endtex 

   !! 0.2 Output variables
   !! notice that these variables (absorption + transmission + reflection) do not necessarily add up to 1 due to multiple scattering
   REAL(r_std), INTENT(OUT)   :: Collim_Alb_Tot              !! collimated total albedo from this level, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(OUT)   :: Collim_Tran_Tot             !! collimated total transmission through this level, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(OUT)   :: Collim_Abs_Tot              !! collimated total absorption by this level, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(OUT)   :: Isotrop_Alb_Tot             !! isotropic total albedo from this level, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(OUT)   :: Isotrop_Tran_Tot            !! isotropic total transmission through this level, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(OUT)   :: Isotrop_Abs_Tot             !! isotropic total absorption by this level, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(OUT)   :: Collim_Tran_Uncollided      !! collimated uncollied transmission through this level, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(OUT)   :: Isotrop_Tran_Uncollided     !! isotropic uncollied transmission through this level, between 0 and 1
                                                             !! @tex $()$ @endtex 

   !! 0.3 Modified variables

   !! 0.4 Local variables
   LOGICAL :: ok
   REAL(r_std), DIMENSION(4)                           :: gammaCoeffs
   REAL(r_std), DIMENSION(4)                           :: gammaCoeffs_star
   REAL(r_std)                                         :: tauprimetilde
   REAL(r_std)                                         :: tauprimestar
   REAL(r_std)                                         :: sun_zenith_angle_radians
   REAL(r_std), PARAMETER                              :: isotropic_cosine_constant=0.5/0.705

   !calculated fluxes

   REAL(r_std)                :: Collim_Alb_BB 
   REAL(r_std)                :: Collim_Tran_BB
   REAL(r_std)                :: Collim_Abs_BB
   REAL(r_std)                :: Isotrop_Alb_BB
   REAL(r_std)                :: Isotrop_Tran_BB
   REAL(r_std)                :: Isotrop_Abs_BB
   REAL(r_std)                :: Collim_Tran_BC
   REAL(r_std)                :: Isotrop_Tran_BC
   REAL(r_std)                :: Collim_Tran_TotalOneWay
   REAL(r_std)                :: Isotrop_Tran_TotalOneWay
   REAL(r_std)                :: Below_reinject_rad_collim,Below_reinject_rad_isotrop

!_ ================================================================================================================================

   IF (bavard.GE.2) WRITE(numout,*) 'Entering twostream_solver'

   ! convert angular values
   
   sun_zenith_angle_radians = acos(cosine_sun_angle)

   ! calculate the 4 gamma coefficients both in isotropic and collimated illumination 

   call gammas(leaf_reflectance, leaf_transmittance, cosine_sun_angle, gammaCoeffs)
   call gammas(leaf_reflectance,leaf_transmittance,isotropic_cosine_constant,&
        gammaCoeffs_star)

   ! estimate the effective value of the optical thickness 

   tauprimetilde = 0.5_r_std * laieff_collim
   !tauprimetilde = 1.

   ! Should become zenit angle dependent (=60°) calculated by the Monte Carlo 
   ! photon shooting routine

   tauprimestar  = 0.5_r_std * laieff_isotrop
   !tauprimestar  = 1.

   ! +++++++++++++ BLACK BACKGROUND ++++++++++++++++++++++ 
   ! * Apply the black-background 2stream solution 
   ! * These equations are written for the part of the incoming radiation 
   ! * that never hits the background but does interact with the vegetation 
   ! * canopy.
   ! * 
   ! * Note : the same routine dhrT1() is used for both the isotropic and
   ! * collimated illumination conditions but the calling arguments differ.
   ! * (especially the solar angle used).
   ! */

   !	/* 1) collimated source */
   ok=dhrT1(leaf_reflectance,leaf_transmittance,&
        gammaCoeffs(1),gammaCoeffs(2),gammaCoeffs(3),gammaCoeffs(4),&
        sun_zenith_angle_radians, tauprimetilde,&
        Collim_Alb_BB,Collim_Tran_BB,Collim_Abs_BB)                      
   ! 	/* 2) isotropic source */
   ok=dhrT1(leaf_reflectance,leaf_transmittance,&
        gammaCoeffs_star(1),gammaCoeffs_star(2),gammaCoeffs_star(3),&
        gammaCoeffs_star(4),&
        acos(isotropic_cosine_constant),tauprimestar,&
        Isotrop_Alb_BB,Isotrop_Tran_BB,Isotrop_Abs_BB)


   ! +++++++++++++ BLACK CANOPY ++++++++++++++++++++++ 
   ! * Apply the black canopy solution.
   ! * These equations hold for the part of the incoming radiation 
   ! * that do not interact with the vegetation, travelling through 
   ! * its gaps.
   ! */

   ! 	/* 1) collimated source */
   IF (cosine_sun_angle .NE. 0) THEN
      Collim_Tran_BC  = exp( - tauprimetilde/cosine_sun_angle)
      Collim_Tran_Uncollided = Collim_Tran_BC
   ENDIF
   ! 	/* 2) isotropic source */
   Isotrop_Tran_BC = TBarreUncoll_exact(tauprimestar)
   Isotrop_Tran_Uncollided = Isotrop_Tran_BC

   ! /* Total one-way transmissions:
   ! * The vegetation canopy is crossed (one way) by the uncollided radiation
   ! * (black canopy) and the collided one (black background). */

   !	/* 1) collimated source */
   Collim_Tran_TotalOneWay  = Collim_Tran_BC  + Collim_Tran_BB 

   !	/* 2) isotropic source */
   Isotrop_Tran_TotalOneWay = Isotrop_Tran_BC + Isotrop_Tran_BB 

   ! * The Below_reinject_rad describes the process of reflecting toward the 
   ! * background the upward travelling radiation (re-emitted from below the 
   ! * canopy). It appears in the coupling equations as the limit of the series: 
   ! *    1 + rg*rbv + (rg*rbv)^2 + (rg*rbv)^3 + ...
   ! *      where rg is the background_reflectance and rbv is Isotrop_Alb_BB
   ! *      (with Isotrop describing the Lambertian reflectance of the background).
   ! */
   ! this might be involved in Eq. 27, 
   Below_reinject_rad_collim = un / (un - background_reflectance_collim*Isotrop_Alb_BB)
   Below_reinject_rad_isotrop = un / (un - background_reflectance_isotrop*Isotrop_Alb_BB)

   !/* TOTAL ALBEDO */
   ! 	/* 1) collimated source */
   Collim_Alb_Tot = Collim_Alb_BB + &
        background_reflectance_collim * Collim_Tran_TotalOneWay * &
        Isotrop_Tran_TotalOneWay * Below_reinject_rad_collim
   !	/* 2) isotropic source */
   Isotrop_Alb_Tot = Isotrop_Alb_BB + & 
        background_reflectance_isotrop * Isotrop_Tran_TotalOneWay * &
        Isotrop_Tran_TotalOneWay * Below_reinject_rad_isotrop ! this seems to be
   ! exactly Eq. 33

   !/* TOTAL TRANSMITION TO THE BACKGROUND LEVEL */
   !	/* 1) collimated source */
   Collim_Tran_Tot = Collim_Tran_TotalOneWay * Below_reinject_rad_collim ;

   !	/* 2) isotropic source */
   Isotrop_Tran_Tot = Isotrop_Tran_TotalOneWay * Below_reinject_rad_isotrop ;

   !/* TOTAL ABSORPTION BY THE VEGETATION LEVEL */
   !	/* 1) collimated source */
   Collim_Abs_Tot = un - (Collim_Tran_Tot + Collim_Alb_Tot) + &
        background_reflectance_collim * Collim_Tran_Tot;
   !	/* 2) isotropic source */
   Isotrop_Abs_Tot = un - (Isotrop_Tran_Tot + Isotrop_Alb_Tot) + &
        background_reflectance_isotrop * Isotrop_Tran_Tot;

   IF (bavard.GE.2) WRITE(numout,*) 'Exiting twostream_solver'

 END SUBROUTINE twostream_solver


!! ============================================================================\n
!! FUNCTION     : dhrT1
!!
!>\BREF         
!!
!! DESCRIPTION : Taken directly from the Fortan code provided by Pinty et al for their \n
!!               two-stream model.  The only changes are the REAL types.  This is used
!!               to compute the black background absorption, transmission, and reflectance
!!               in Pintys scheme and it's based on the older model of Meador and Weaver...
!!               Pinty 2006 also includes a short discussion of it in Appendix A 
!!
!! RECENT CHANGE(S): None\n
!!
!! RETURN VALUE : dhrT1
!!
!! REFERENCE(S) : Meador and Weaver, 'Two-stream approximations to radiative transfer in
!!    planetary atmosphere: a unified description of existing methods and a new improvement'
!!    J. Atmospheric Sciences, VOL 37, p. 630--643, 1980.
!!
!! FLOWCHART    : None
!_ =============================================================================

FUNCTION dhrT1(rl,tl,gamma1,gamma2,gamma3,gamma4,tta,tau,AlbBS,Tdif,AbsVgt)


  !! 0. Variables and parameter declaration
  !! 0.1 Input variables
  REAL(r_std), INTENT(IN) :: rl  ! the effective reflectance of a single scatterer, between 0 and 1
                                 !! @tex $()$ @endtex      
  REAL(r_std), INTENT(IN) :: tl  ! the effective transmittance of a single scatterer, between 0 and 1
                                 !! @tex $()$ @endtex      
  REAL(r_std), INTENT(IN) :: gamma1 ! the gamma coefficients from Meador and Weaver
  REAL(r_std), INTENT(IN) :: gamma2
  REAL(r_std), INTENT(IN) :: gamma3
  REAL(r_std), INTENT(IN) :: gamma4
  REAL(r_std), INTENT(IN) :: tta ! the solar zenith angle, between 0 and pi/2
                                 !! @tex $(radians)$ @endtex      

  REAL(r_std), INTENT(IN) :: tau ! Effective LAI * G(theta)
                                 !! @tex $()$ @endtex      

  !! 0.2 Output variables
  REAL(r_std), INTENT(OUT) :: AlbBS ! the albedo of level, between 0 and 1
                                 !! @tex $()$ @endtex      
  REAL(r_std), INTENT(OUT) :: Tdif ! the transmitted light (all diffuse) from this light source, between 0 and 1
                                 !! @tex $()$ @endtex      
  REAL(r_std), INTENT(OUT) :: AbsVgt ! the light absorbed by the level, between 0 and 1
                                 !! @tex $()$ @endtex      
  LOGICAL :: dhrT1 ! the output flag, which always appears to be true

  !! 0.3 Modified variables
  !! 0.4 Local variables

  REAL(r_std) :: alpha1,alpha2,ksquare,k
  REAL(r_std) :: first_term,secnd_term1,secnd_term2,secnd_term3
  REAL(r_std) :: expktau,Tdir
  REAL(r_std) :: mu0,w0 

  !_ ================================================================================================================================

  mu0=cos(tta)
  w0=rl+tl


  Tdir = exp(-tau/mu0) ! direct transmission

  ! There is a difference between conservative and non-conservative scattering 
  ! conditions */
  IF (w0 .ne. 1.0 .AND. w0 .ne. 0.0) THEN
     !NON_CONSERVATIVE SCATTERING

     ! some additional parameters, taken from Eq. 16--18 of Meador and Weaver, J. Atmospheric Sciences,
     ! Vol 37, p. 630--643 (1980)
     alpha1  = gamma1*gamma4 + gamma2*gamma3
     alpha2  = gamma1*gamma3 + gamma2*gamma4
     ksquare = gamma1*gamma1 - gamma2*gamma2
     k       = sqrt(ksquare)

     expktau = exp(k*tau)

     !Black Soil Albedo...Eq. 14 in Meador and Weaver
     first_term  = ((un-ksquare*mu0*mu0)*((k+gamma1)*expktau + (k-gamma1)/expktau))
     IF (first_term .eq. 0.0) THEN
        !we will be dividing by zero : cannot continue.
        dhrT1 = .false.
     ELSE 
        first_term = un/first_term
        secnd_term1 = (un - k*mu0)*(alpha2 + k*gamma3)*expktau
        secnd_term2 = (un + k*mu0)*(alpha2 - k*gamma3)/expktau
        secnd_term3 = 2.0_r_std * k * (gamma3 - alpha2*mu0)*Tdir
        AlbBS = (w0 * first_term * (secnd_term1 - secnd_term2 - secnd_term3))

        !Transmission...Eq. 15 in Meador and Weaver, for diffuse light?
        IF (ksquare .eq. 0.0) THEN 
           first_term = un
        ENDIF
        secnd_term1 = (un+k*mu0)*(alpha1+k*gamma4)*expktau
        secnd_term2 = (un-k*mu0)*(alpha1-k*gamma4)/expktau
        secnd_term3 = 2.0_r_std * k * (gamma4 + alpha1*mu0)
        Tdif = - w0*first_term*(Tdir*(secnd_term1 - secnd_term2) - secnd_term3)

        ! Absorption by vegetation...whatever is not transmitted or reflected 
        ! must be absorbed
        AbsVgt = (un- (Tdif+Tdir) - AlbBS)
     ENDIF ! first_term .eq. 0.0
  ELSE IF (w0 .eq. 0.) THEN
     !BLACK CANOPY
     AlbBS = zero
     Tdif  = zero
     AbsVgt = un - Tdir
  ELSE
     !CONSERATIVE SCATTERING...Eq. 24 in Meador and Weaver
     AlbBS =  (un/(un + gamma1*tau))*(gamma1*tau + (gamma3-gamma1*mu0)*&
          (un-exp(-tau/mu0)));
     Tdif   = un - AlbBS - Tdir;
     AbsVgt = zero;
  ENDIF ! w0 .ne. 1.0 .AND. w0 .ne. 0.0

! not sure what the purpose of this flag is, as it will always be set to true here
  dhrT1 = .true.

END FUNCTION dhrT1

!! ============================================================================\n
!! FUNCTION     : TBarreUncoll_exact
!!
!>\BRIEF          Computes the transmission of the diffuse black canopy light
!!
!! DESCRIPTION : Taken directly from the Fortan code provided by Pinty et al for their \n
!!               two-stream model.  The only changes are the REAL types.  This appears
!!               to be solving Eq. 16 in Pinty 2006, which is the transmission which
!!               does not collide with the canopy
!!
!! RECENT CHANGE(S): None\n
!!
!! RETURN VALUE : TBarreUncoll_exact
!!
!! REFERENCE(S) : None
!!
!! FLOWCHART    : None
!_ =============================================================================

FUNCTION TBarreUncoll_exact(tau)


  !! 0. Variables and parameter declaration

  !! 0.1 Input variables
  REAL(r_std), INTENT(IN) :: tau ! Effective LAI * G(theta)
  !! @tex $()$ @endtex      

  !! 0.2 Output variables
  REAL(r_std) :: TBarreUncoll_exact ! the isotropic light transmission uncollided
                                    ! with the canopy, between 0 and 1
  !! @tex $()$ @endtex      


  !! 0.3 Modified variables

  !! 0.4 Local variables
!  INTEGER :: j,ind
!  REAL(r_std) :: iGammaloc
!  INTEGER :: order

!_ ================================================================================================================================

  !+++++ CHECK +++++

!!$  iGammaloc = zero
!!$  order=20
!!$
!!$  ! in the case where tau is equal to zero, this crashes in the first loop where
!!$  ! iGammaloc is also zero...quick fix, and might even be accurate
!!$
!!$  IF(tau .GT. 1e-10_r_std)THEN
!!$
!!$     DO j=0,order-1
!!$        ind = order - j
!!$        iGammaloc = ind / (un + ind/(tau+iGammaloc))
!!$     END DO
!!$     iGammaloc=un/(tau + iGammaloc)
!!$  ENDIF
!!$
!!$  TBarreUncoll_exact = exp(-tau)*(un - tau + tau*tau*iGammaloc)

  ! this is a change suggested by Bernard to improve the matching between the one
  ! level and multilevel case

  TBarreUncoll_exact = exp(-tau*2.0_r_std*0.705_r_std)
  !++++++++++

END FUNCTION TBarreUncoll_exact

!! ==============================================================================================================================
!! SUBROUTINE   : gammas
!!
!>\BRIEF         Computes a set of gamma coefficients for use in the black background equations
!!
!! DESCRIPTION : Taken directly from the Fortan code provided by Pinty et al for their \n
!!               two-stream model.  The only changes are the REAL types.  This calculates
!!               the gamma coefficients based on Appendix A in the paper.  This seems to
!!               make the assumption of the Ross's G function and a spherical leaf
!!               angle distribution.
!!
!! RECENT CHANGE(S): None\n
!!
!! MAIN OUTPUT VARIABLE(S): gammaCoeff
!!
!! REFERENCE(S) :  B. Pinty, T. Lavergne, R.E. Dickinson, J-L. Widlowski, N. Gobron and M. M. Verstraete (2006).
!! Simplifying the Interaction of Land Surfaces with Radiation for Relating Remote Sensing Products to Climate Models.
!! Journal of Geophysical Research. Vol 111, D02116.
!!
!! FLOWCHART    : None
!_ ================================================================================================================================

SUBROUTINE gammas(rl, tl, mu0, gammaCoeff)


  !! 0. Variables and parameter declaration
  !! 0.1 Input variables
  REAL(r_std), INTENT(IN) :: rl  ! the effective reflectance of a single scatterer, between 0 and 1
  !! @tex $()$ @endtex      
  REAL(r_std), INTENT(IN) :: tl  ! the effective transmittance of a single scatterer, between 0 and 1
  !! @tex $()$ @endtex      
  REAL(r_std), INTENT(IN) :: mu0 ! the cosine of the solar zenith angle, between 0 and 1
  !! @tex $()$ @endtex      

  !! 0.2 Output variables
  REAL(r_std), DIMENSION(1:4), INTENT(OUT) :: gammaCoeff ! the set of gamma coefficients in the above reference
  !! @tex $()$ @endtex      

  !! 0.3 Modified variables

  !! 0.4 Local variables
  REAL(r_std) :: wd,w0,w0half,wdsixth

!_ ================================================================================================================================


  w0      = rl+tl
  wd      = rl-tl
  w0half  = w0*0.5_r_std
  wdsixth = wd/6.0_r_std

  gammaCoeff(1)=2._r_std*(un - w0half + wdsixth)
  gammaCoeff(2)=2._r_std*(w0half + wdsixth)
  gammaCoeff(3)=2._r_std*(w0half*0.5_r_std + mu0*wdsixth)/w0
  gammaCoeff(4)=un-gammaCoeff(3)

END SUBROUTINE gammas


!! ==============================================================================================================================
!! SUBROUTINE   : calculate_snow_albedo
!!
!>\BRIEF         Computes some of the information needed to calculate the effect of the snow albedo
!!               on the background reflectance in the two stream model.  Right now, this can be done
!!               with the old snow albedo scheme from Krinner et al 2005 as well as the snow albedo
!!               scheme from Community Land Model 3 (CLM3). The calculation of
!!               snow cover fraction is taken from Yang et al. 1997
!!
!! DESCRIPTION : In order to compute the background albedo in Pinty's two stream model, we have to take
!!               into account any snow that has fallen through the canopy and landed on the ground.  In particular,
!!               we need the amount of ground covered by the snow and the albedo of this snow.  Both of these
!!               quantities are calculated here, using a function that changes the albedo of the snow based on
!!               its age.  This is done in two ways, but the model of CLM3 is better suited to be used with Pinty's
!!               albedo scheme, as it divides the radiation into VIS and NIR light. The calculation of snow cover 
!!               fraction depends on roughness lenght.
!!
!! RECENT CHANGE(S): None\n
!!
!! MAIN OUTPUT VARIABLE(S): ::snowa_veg, ::frac_snow_veg, ::albedo_snow
!!
!! REFERENCE(S) :  "Technical description of the community land model CLM)", K. W. Oleson, Y. Dai, G. Bonan, M. Bosilovich, et al.,
!!                 NCAR Technical Note, May 2004.
!!
!!                 Yang Z, Dickinson R, Robock A, Vinnikov K (1997) Validation of the snow submodel of the 
!!                 biosphere-atmosphere transfer scheme with Russian 
!!                 snow cover and meteorological observational data. Journal of Climate, 10, 353–373.
!!
!! FLOWCHART    : None
!_ ================================================================================================================================

SUBROUTINE calculate_snow_albedo(npts,  sinang,  snow,  snow_nobio, &
           snow_age,  snow_nobio_age,  frac_nobio,  albedo_snow, &
           snowa_veg,  frac_snow_veg,  snowa_nobio,  frac_snow_nobio, &
           veget_max, z0_veg)

  !! 0. Variables and parameter declaration
  !! 0.1 Input variables
  INTEGER,INTENT(IN)          :: npts !! Domain size - Number of land pixels  (unitless) 
  REAL(r_std), DIMENSION(npts), INTENT(in)      :: sinang        !!  cosine of the solar zenith angle, between 0 and 1
                                                                     !! @tex $()$ @endtex 
   REAL(r_std),DIMENSION (npts), INTENT(in)        :: snow            !! Snow mass in vegetation (kg m^{-2})           
   REAL(r_std),DIMENSION (npts,nnobio), INTENT(in) :: snow_nobio      !! Snow mass on continental ice, lakes, etc. (kg m^{-2})      
   REAL(r_std),DIMENSION (npts), INTENT(in)        :: snow_age        !! Snow age (days)        
   REAL(r_std),DIMENSION (npts,nnobio), INTENT(in) :: snow_nobio_age  !! Snow age on continental ice, lakes, etc. (days)    
   REAL(r_std),DIMENSION (npts,nnobio), INTENT(in) :: frac_nobio      !! Fraction of non-vegetative surfaces, i.e. 
                                                                          !! continental ice, lakes, etc. (unitless)     
   REAL(r_std), DIMENSION(npts,nvm), INTENT(in)    :: veget_max       !! PFT coverage fraction of a PFT (= ind*cn_ind) 
                                                                          !! (m^2 m^{-2})
   REAL(r_std), DIMENSION(npts), INTENT(in)        :: z0_veg          !! Roughness height of vegetated part (m)


  !! 0.2 Output variables
  REAL(r_std), DIMENSION(npts),INTENT(OUT) :: albedo_snow           !! the averaged snow albedo for a given grid cell
  REAL(r_std), DIMENSION(npts,nvm,n_spectralbands),INTENT(OUT) :: snowa_veg  !! the albedo of snow...this is seperated into
                                                                                 !! PFT types, even though the calculation is identical
                                                                                 !! for all PFTs right now
  REAL(r_std), DIMENSION(npts,nvm),INTENT(OUT) :: frac_snow_veg              !! the fraction of the surface for each PFT covered by snow
  REAL(r_std), DIMENSION(npts,nnobio,n_spectralbands),INTENT(OUT) :: snowa_nobio !! the albedo of snow on nonbiological land types
  REAL(r_std), DIMENSION(npts,nnobio),INTENT(OUT) :: frac_snow_nobio    !! the fraction of nonbiological land types covered by snow


  !! 0.3 Modified variables

  !! 0.4 Local variables
  REAL(r_std) :: agefunc,agefunc_nobio,snow_alb_direct, snow_alb_diffuse,f_mu
  REAL(r_std),DIMENSION(n_spectralbands)            :: c_albedo,alb_snow_0 ! parameter values, taken directly from CLM3

  INTEGER :: ipts,ks,ivm,jv ! indices
  
  !_ ================================================================================================================================

  
  ! initialize the output
  albedo_snow=val_exp
  snowa_veg=val_exp
  frac_snow_veg=val_exp
  snowa_nobio=val_exp
  frac_snow_nobio=val_exp  

  ! set some values for the new snow albedo
  alb_snow_0(ivis)=0.95_r_std
  alb_snow_0(inir)=0.65_r_std
  c_albedo(ivis)=0.2_r_std
  c_albedo(inir)=0.5_r_std


  DO ipts = 1, npts ! loop over all the grid squares


     ! calculate the snow cover fraction...right now this doesn't depend on PFT
     ! snow fraction calculated following Yang et al 1997, eq. 32  
     frac_snow_veg(ipts,:) = TANH((snow(ipts)/sn_dens)/ &
          MAX((2.5_r_std*z0_veg(ipts)), &
          min_sechiba)) 

     DO ivm=1,nvm ! loop over all the PFTs


        IF(veget_max(ipts,ivm) == zero)THEN
            ! this vegetation type is not present, so no reason to do the 
            ! calculation
            CYCLE
        ENDIF


        DO ks=1,n_spectralbands ! loop over spectra

           IF (ABS(fixed_snow_albedo - undef_sechiba) .GT. &
                EPSILON(undef_sechiba)) THEN
              snowa_veg(ipts,ivm,ks) = fixed_snow_albedo
           ELSE

              IF(control%do_new_snow_albedo)THEN
                 ! first we calculate diffuse albedo
                 ! NOTE: The difference between these two methods has not been 
                 ! tested.  The only constraint is that it must be dimensionless
                 agefunc = un - EXP(-snow_age(ipts)/tcst_snowa) ! already in ORCHIDEE
                 !agefunc = zero

                 ! using the constant from CLM3 conveted into days**-1
                 !agefunc=un-un/(un+snow_age(ipts)*0.864_r_std)
                 !agefunc=un/(un+snow_age(ipts)*0.864_r_std)     
                 snow_alb_diffuse=(un-c_albedo(ks)*agefunc)*alb_snow_0(ks)

                 ! now the direct albedo
                 IF(sinang(ipts) .GT. 0.5_r_std)THEN
                    f_mu=zero
                 ELSE
                    ! substituting b=2.0 into the equation...recommended by CLM3
                    f_mu=(3.0_r_std)/(un+sinang(ipts))-2.0_r_std 
                 ENDIF

                 snow_alb_direct=snow_alb_diffuse+0.4_r_std*f_mu*&
                      (un-snow_alb_diffuse)

                 ! assume that the total snow albedo is a mix of 
                 ! 50% direct and 50% diffuse
                 snowa_veg(ipts,ivm,ks)=(snow_alb_direct+snow_alb_diffuse)&
                      /2.0_r_std

              ELSE
                 ! calculate the age of the snow for this grid square
                 agefunc = EXP(-snow_age(ipts)/tcst_snowa)

                 ! now the albedo...use only the bare soil snow decay parameters, 
                 ! since the true albedo is calculated
                 ! by the PFT-dependent two-stream solver
                 snowa_veg(ipts,ivm,ks) = snowa_aged(1)+snowa_dec(1)*agefunc 

              ENDIF ! control%do_new_snow_albedo
           ENDIF ! prescribe or calculate albedo



           ! Notice that the fraction of nobio land is included in veget_max.
           ! We have to average by the number of spectras, too, for this array, 
           ! because it doesn't distinguish between PFTs or spectra
           albedo_snow(ipts) = albedo_snow(ipts) + &
                frac_snow_veg(ipts,ivm) * veget_max(ipts,ivm) * &
                snowa_veg(ipts,ivm,ks)/REAL(n_spectralbands,r_std)

        ENDDO ! ks=1,n_spectralbands 

     ENDDO ! ivm=1,nvm 

     ! there is no distinction between NIR and VIS for non bio surfaces
     IF (ABS(fixed_snow_albedo - undef_sechiba) .GT. EPSILON(undef_sechiba)) THEN
        snowa_nobio(:,:,:) = fixed_snow_albedo
     ELSE

        DO jv = 1, nnobio 
           ! calculate the snow age
           agefunc_nobio = EXP(-snow_nobio_age(ipts,jv)/tcst_snowa)

           ! and the albedo due to snow of this age on this nonbio surface
           snowa_nobio(ipts,jv,:) = ( snowa_aged(1) + snowa_dec(1) * agefunc_nobio ) 
        ENDDO
     ENDIF ! prescribe or calculate

     DO jv = 1, nnobio 
        ! calculate the fraction of the surface covered by snow
        ! snow fraction calculated following Yang et al 1997, eq. 32
        frac_snow_nobio(ipts,jv) = TANH((snow_nobio(ipts,jv)/sn_dens)/ &
      MAX((2.5_r_std*z0_ice), &
             min_sechiba))

        ! For some reason, occasionally snow_nobio is negative.  This is driver
        ! data that we cannot control, but it should not influence our snowcover, so
        ! set it equal to zero in that case.
        IF(snow_nobio(ipts,jv) .LT. zero) frac_snow_nobio(ipts,jv)=zero

        albedo_snow(ipts) = albedo_snow(ipts) + &
             ( frac_nobio(ipts,jv) ) * ( frac_snow_nobio(ipts,jv) ) * &
             snowa_nobio(ipts,jv,1) ! not sensitive to the spectral band.  
        ! Don't need to average here since we are outside the loop over the bands

     ENDDO ! jv = 1, nnobio 
     
  ENDDO ! ipts = 1, npts 




END SUBROUTINE CALCULATE_SNOW_ALBEDO

!! ==============================================================================================================================
!! SUBROUTINE   : optimize_albedo_values
!!
!>\BRIEF         Follow the radiation scattered through the canopy in the 
!!               case of multiple levels.
!!
!! DESCRIPTION : Right now, we know that using Pintys method in the case of a 
!!               single canopy level will result in different top of the canopy
!!               albedos than iwhat is found if we use multiple canopy levels.
!!               We trust that the single level case gives the 'true' values.
!!
!!               This algorithm follows light as it scatters through multiple
!!               levels in the canopy.  At each level, the scattering 
!!               solution is found by solving Pinty's two stream model.  This
!!               means that light which enters a level can either be transmitted
!!               without colliding with the vegetation, transmitted after
!!               collision with the vegetation, reflecting off the vegetation,
!!               or being absorbed.  We follow the fluxes until they get
!!               really small.
!!
!! RECENT CHANGE(S): None\n
!!
!! MAIN OUTPUT VARIABLE(S): ::lconverged, ::Collim_Alb_Tot, ::Collim_Tran_Tot, 
!!          ::Collim_Abs_Tot, ::Isotrop_Alb_Tot, ::Isotrop_Tran_Tot, ::Isotrop_Abs_Tot
!!
!! REFERENCE(S) :  B. Pinty, T. Lavergne, R.E. Dickinson, J-L. Widlowski, N. Gobron
!!          and M. M. Verstraete (2006). Simplifying the Interaction of Land 
!!          Surfaces with Radiation for Relating Remote Sensing Products to 
!!          Climate Models. Journal of Geophysical Research. Vol 111, D02116.
!!
!! FLOWCHART    : None
!_ ================================================================================================================================
SUBROUTINE multilevel_albedo(nlevels_loc, cosine_sun_angle, leaf_single_scattering_albedo_start,&
     leaf_psd_start, br_base_temp_collim,  br_base_temp_isotrop, &
     laieff_collim_pft, laieff_isotrop_pft, lconverged, &
     Collim_Alb_Coll, Collim_Tran_Coll, Collim_Abs_Tot, Isotrop_Alb_Coll, &
     Isotrop_Tran_Coll, Isotrop_Abs_Tot, Collim_Tran_Uncoll, Isotrop_Tran_Uncoll, lprint)

  !! 0. Variables and parameter declaration
  !! 0.1 Input variables
  INTEGER,INTENT(IN)        :: nlevels_loc 
  REAL(r_std), INTENT(IN)    :: cosine_sun_angle            !! cosine of the solar zenith angle, between 0 and 1
  !! @tex $()$ @endtex 
  REAL(r_std), INTENT(IN)    :: leaf_single_scattering_albedo_start           !! cosine of the solar zenith angle, between 0 and 1
  REAL(r_std), INTENT(IN)    :: leaf_psd_start            !! cosine of the solar zenith angle, between 0 and 1
  REAL(r_std), INTENT(IN)    :: br_base_temp_collim           !! cosine of the solar zenith angle, between 0 and 1
  REAL(r_std), INTENT(IN)    :: br_base_temp_isotrop            !! cosine of the solar zenith angle, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(IN)        :: laieff_collim_pft                 !! Effective lai for a single pixel and pft to be 
  REAL(r_std),DIMENSION(:),INTENT(IN)        :: laieff_isotrop_pft                 !! Effective lai for a single pixel and pft to be 
  LOGICAL,INTENT(IN)        :: lprint                 !! A flag to print some debug statements
  !!  used in the two stream approach...every level

  !! 0.2 Output variables
  LOGICAL,INTENT(OUT) :: lconverged                     !! did the optimization converge?
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Collim_Alb_Coll   !! collimated total albedo for the converged case
  !! unitless, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Collim_Tran_Coll  !! collimated total transmission
  !! unitless, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Collim_Abs_Tot   !! collimated total absorption
  !! unitless, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Isotrop_Alb_Coll  !! isotropic total albedo
  !! unitless, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Isotrop_Tran_Coll !! isotropic total transmission
  !! unitless, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Isotrop_Abs_Tot  !! isotropic total absorption
  !! unitless, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Collim_Tran_Uncoll  !! collimated uncollided transmission
  !! unitless, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Isotrop_Tran_Uncoll  !! isotropic uncollided transmission
  !! unitless, between 0 and 1

  !! 0.3 Modified variables

  !! 0.4 Local variables
  ! use an extra level here...this is basically to store the transmission from the sun, which is normalized to 1.0
  REAL(r_std),DIMENSION(nlevels_loc)            :: Collim_Abs_Coll_Unscaled
  REAL(r_std),DIMENSION(nlevels_loc)            :: Collim_Alb_Coll_Unscaled
  REAL(r_std),DIMENSION(nlevels_loc)            :: Collim_Tran_Coll_Unscaled
  REAL(r_std),DIMENSION(nlevels_loc)            :: Collim_Tran_UnColl_Unscaled
  REAL(r_std),DIMENSION(nlevels_loc)            :: Isotrop_Abs_Coll_Unscaled
  REAL(r_std),DIMENSION(nlevels_loc)            :: Isotrop_Alb_Coll_Unscaled
  REAL(r_std),DIMENSION(nlevels_loc)            :: Isotrop_Tran_Coll_Unscaled
  REAL(r_std),DIMENSION(nlevels_loc)            :: Isotrop_Tran_UnColl_Unscaled
  REAL(r_std),DIMENSION(nlevels_loc)            :: Collim_Tran_Tot
  REAL(r_std),DIMENSION(nlevels_loc)            :: Isotrop_Tran_Tot

  REAL(r_std),DIMENSION(0:nlevels_loc+1)            :: &
       collim_down_cs,isotrop_down_cs,isotrop_up_cs

  INTEGER                                   :: istep,ilevel

  REAL(r_std)                               :: leaf_reflectance
  REAL(r_std)                               :: leaf_transmittance

  LOGICAL :: lexit
  LOGICAL :: ok
  REAL(r_std), DIMENSION(4)                           :: gammaCoeffs
  REAL(r_std), DIMENSION(4)                           :: gammaCoeffs_star
  REAL(r_std)                                         :: sun_zenith_angle_radians
  REAL(r_std), PARAMETER                              :: isotropic_cosine_constant=0.5/0.705

  INTEGER,SAVE :: icalls=0
  !_ ================================================================================================================================

  
  icalls=icalls+1

  ! convet the sun angle
  sun_zenith_angle_radians = acos(cosine_sun_angle)



  ! initialize some values that are used in the optimization loop
  istep=0
  lexit=.FALSE.
  lconverged=.FALSE.

  ! calculate the albedo at our starting point

  leaf_transmittance = leaf_single_scattering_albedo_start/ & 
       ( leaf_psd_start+un)
  leaf_reflectance =  leaf_psd_start * &
       leaf_transmittance

  ! some debugging stuff
  !  DO ilevel=1,nlevels_loc

  !     CALL print_debugging_albedo_info(ilevel,cosine_sun_angle,leaf_reflectance,&
  !          leaf_transmittance,&
  !          laieff_collim_pft(ilevel),laieff_isotrop_pft(ilevel), &
  !          reflectance_collim(ilevel-1),reflectance_isotrop(ilevel-1),&
  !          Collim_Alb_Tot_temp(ilevel),Collim_Tran_Tot_temp(ilevel),&
  !          Collim_Abs_Tot_temp(ilevel),&
  !          Isotrop_Alb_Tot_temp(ilevel),Isotrop_Tran_Tot_temp(ilevel),&
  !          Isotrop_Abs_Tot_temp(ilevel))
  !  ENDDO


  ! calculate the gamma coefficients used in the case of the black background 
  call gammas(leaf_reflectance, leaf_transmittance, cosine_sun_angle, gammaCoeffs)
  call gammas(leaf_reflectance,leaf_transmittance,isotropic_cosine_constant,&
       gammaCoeffs_star)



  !************************** step one ****** !
  ! compute all the unscaled quantities
  DO ilevel=nlevels_loc,1,-1

     Collim_Tran_UnColl_Unscaled(ilevel)=exp( - 0.5_r_std*&
          laieff_collim_pft(ilevel)/cosine_sun_angle)

     Isotrop_Tran_UnColl_Unscaled(ilevel)= &
          TBarreUncoll_exact(0.5_r_std*laieff_isotrop_pft(ilevel))


     !	/* 1) collimated source */
     ok=dhrT1(leaf_reflectance,leaf_transmittance,&
          gammaCoeffs(1),gammaCoeffs(2),gammaCoeffs(3),gammaCoeffs(4),&
          sun_zenith_angle_radians, 0.5_r_std*laieff_collim_pft(ilevel),&
          Collim_Alb_Coll_Unscaled(ilevel),Collim_Tran_Coll_Unscaled(ilevel),&
          Collim_Abs_Coll_Unscaled(ilevel))

     ! 	/* 2) isotropic source */
     ok=dhrT1(leaf_reflectance,leaf_transmittance,&
          gammaCoeffs_star(1),gammaCoeffs_star(2),gammaCoeffs_star(3),gammaCoeffs_star(4),&
          acos(isotropic_cosine_constant),0.5_r_std*laieff_isotrop_pft(ilevel),&
          Isotrop_Alb_Coll_Unscaled(ilevel),Isotrop_Tran_Coll_Unscaled(ilevel),&
          Isotrop_Abs_Coll_Unscaled(ilevel))

     !     Collim_Tran_OneWay_Unscaled(ilevel)=Collim_Tran_Coll_Unscaled(ilevel)+&
     !              Collim_Tran_UnColl_Unscaled(ilevel)
     !     Isotrop_Tran_OneWay_Unscaled(ilevel)=Isotrop_Tran_Coll_Unscaled(ilevel)+&
     !              Isotrop_Tran_UnColl_Unscaled(ilevel)

     !     WRITE(6,'(A,I5,3F10.6)') 'unscaled ',ilevel,Collim_Tran_UnColl_Unscaled(ilevel),&
     !          Collim_Tran_Coll_Unscaled(ilevel),&
     !          Collim_Alb_Coll_Unscaled(ilevel)
  ENDDO ! ilevel=nlevels_loc,1,-1




  ! Following the fate of the light at every step

  ! The downwelling array indicates the quantity of light flowing into this level
  ! from above therefore, to start our system for collimated light, we give a 
  ! unit of light coming into the top level from a collimated source.
  ! The upwelling array is light entering this level from below.
  ! cs is the current step, while ns is the next step for the iteration.
  ! Level 0 is the background. Light can enter this level from above but not below
  ! Level nlevels_loc+1 is the atomosphere. Light can enter this level from below 
  ! but not above.
  collim_down_cs(:)=zero
  collim_down_cs(nlevels_loc)=un
  isotrop_down_cs(:)=zero
  isotrop_up_cs(:)=zero

  CALL propagate_fluxes(nlevels_loc, collim_down_cs, isotrop_down_cs, isotrop_up_cs, &
       Collim_Tran_UnColl_Unscaled, Collim_Tran_Coll_Unscaled, &
       Collim_Alb_Coll_Unscaled, Isotrop_Tran_UnColl_Unscaled, &
       Isotrop_Tran_Coll_Unscaled, Isotrop_Alb_Coll_Unscaled, &
       br_base_temp_collim, br_base_temp_isotrop, .FALSE., &
       Collim_Tran_Uncoll, Collim_Tran_Coll, Collim_Alb_Coll, lconverged, lprint)

  ! Now for the isotropic light
  collim_down_cs(:)=zero
  isotrop_down_cs(:)=zero
  isotrop_down_cs(nlevels_loc)=un
  isotrop_up_cs(:)=zero

  ! The colliminated light is never used here, but it's passed to keep things 
  ! clean between the two cases. Might have to redo this if we are having 
  ! peformance issues.
  CALL propagate_fluxes(nlevels_loc, collim_down_cs, isotrop_down_cs, isotrop_up_cs, &
       Collim_Tran_UnColl_Unscaled, Collim_Tran_Coll_Unscaled, &
       Collim_Alb_Coll_Unscaled, Isotrop_Tran_UnColl_Unscaled, &
       Isotrop_Tran_Coll_Unscaled, Isotrop_Alb_Coll_Unscaled, &
       br_base_temp_collim, br_base_temp_isotrop, .TRUE., &
       Isotrop_Tran_Uncoll, Isotrop_Tran_Coll, Isotrop_Alb_Coll, lconverged, lprint)

  ! Calculate the absorption profile
  Collim_Tran_Tot(:)=Collim_Tran_Coll(:)+Collim_Tran_Uncoll(:)
  Isotrop_Tran_Tot(:)=Isotrop_Tran_Coll(:)+Isotrop_Tran_Uncoll(:)

  ! bottom level
  Collim_Abs_Tot(1)=Collim_Tran_Tot(1+1)+Collim_Tran_Tot(1)*br_base_temp_collim&
       -Collim_Tran_Tot(1)-Collim_Alb_Coll(1)
  Isotrop_Abs_Tot(1)=Isotrop_Tran_Tot(1+1)+Isotrop_Tran_Tot(1)*br_base_temp_isotrop&
       -Isotrop_Tran_Tot(1)-Isotrop_Alb_Coll(1)

  ! all middle levels
  DO ilevel=2,nlevels_loc-1
     Collim_Abs_Tot(ilevel)=Collim_Tran_Tot(ilevel+1)+Collim_Alb_Coll(ilevel-1)&
          -Collim_Tran_Tot(ilevel)-Collim_Alb_Coll(ilevel)
     Isotrop_Abs_Tot(ilevel)=Isotrop_Tran_Tot(ilevel+1)+Isotrop_Alb_Coll(ilevel-1)&
          -Isotrop_Tran_Tot(ilevel)-Isotrop_Alb_Coll(ilevel)
  ENDDO

  ! top level
  Collim_Abs_Tot(nlevels_loc)=un+Collim_Alb_Coll(nlevels_loc-1)&
       -Collim_Tran_Tot(nlevels_loc)-Collim_Alb_Coll(nlevels_loc)
  Isotrop_Abs_Tot(nlevels_loc)=un+Isotrop_Alb_Coll(nlevels_loc-1)&
       -Isotrop_Tran_Tot(nlevels_loc)-Isotrop_Alb_Coll(nlevels_loc)



END SUBROUTINE multilevel_albedo

!! ==============================================================================================================================
!! SUBROUTINE   : print_debugging_albedo_info
!!
!>\BRIEF         Prints out some albedo information in a nice format.  
!!               Should only be used for debugging, never for production runs.
!!
!! DESCRIPTION : 
!!
!! RECENT CHANGE(S): None\n
!!
!! MAIN OUTPUT VARIABLE(S): None.
!!
!! REFERENCE(S) :  
!!
!! FLOWCHART    : None
!_ ================================================================================================================================
SUBROUTINE print_debugging_albedo_info(ilevel, cosine_sun_angle, &
     leaf_reflectance, leaf_transmittance, laieff_collim_temp, laieff_isotrop_temp, &
     br_base_temp_collim, br_base_temp_isotrop, Collim_Alb_Tot, &
     Collim_Tran_Tot, Collim_Abs_Tot, Isotrop_Alb_Tot, Isotrop_Tran_Tot, Isotrop_Abs_Tot)

  !! 0. Variables and parameter declaration
  !! 0.1 Input variables
  INTEGER,INTENT(IN)      :: ilevel
  REAL(r_std), INTENT(IN)    :: cosine_sun_angle            !! cosine of the solar zenith angle, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(IN)    :: laieff_collim_temp           !! cosine of the solar zenith angle, between 0 and 1
   REAL(r_std), INTENT(IN)    :: laieff_isotrop_temp           !! cosine of the solar zenith angle, between 0 and 1
   REAL(r_std), INTENT(IN)    :: leaf_reflectance           !! cosine of the solar zenith angle, between 0 and 1
   REAL(r_std), INTENT(IN)    :: leaf_transmittance            !! cosine of the solar zenith angle, between 0 and 1
   REAL(r_std), INTENT(IN)    :: br_base_temp_collim           !! cosine of the solar zenith angle, between 0 and 1
   REAL(r_std), INTENT(IN)    :: br_base_temp_isotrop            !! cosine of the solar zenith angle, between 0 and 1
   REAL(r_std),INTENT(IN)          :: Collim_Alb_Tot
   REAL(r_std),INTENT(IN)          :: Collim_Tran_Tot
   REAL(r_std),INTENT(IN)          :: Collim_Abs_Tot
   REAL(r_std),INTENT(IN)          :: Isotrop_Alb_Tot
   REAL(r_std),INTENT(IN)          :: Isotrop_Tran_Tot
   REAL(r_std),INTENT(IN)          :: Isotrop_Abs_Tot

  !! 0.2 Output variables

  !! 0.3 Modified variables

  !! 0.4 Local variables

  !_ ================================================================================================================================
   
1223 FORMAT(I5,I3,7(F14.7))

     WRITE (6,'(8(11A))') '   Level   ','   Angle   ','  laieff_c  ','  laieff_i  ',&
          '   rleaf   ','   tleaf   ','   rbgd_c   ','   rbgd_i   '

     WRITE(6,FMT='(I6,5X,7(F11.6))') &
          ilevel,180/pi*ACOS(cosine_sun_angle),&
          laieff_collim_temp,laieff_isotrop_temp,&
          leaf_reflectance,leaf_transmittance,br_base_temp_collim,&
          br_base_temp_isotrop


     WRITE (6,'(10X,6(11A))') ' Rtot(sun) ',' Ttot(sun) ',&
          ' Atot(sun) ',' Rtot(iso) ',' Ttot(iso) ',' Atot(iso) '
     WRITE(6,FMT='(10X,6(F11.6))') &
          Collim_Alb_Tot,Collim_Tran_Tot,Collim_Abs_Tot,&
          Isotrop_Alb_Tot,Isotrop_Tran_Tot,Isotrop_Abs_Tot


END SUBROUTINE print_debugging_albedo_info


!! ==============================================================================================================================
!! SUBROUTINE   : propagate_fluxes
!!
!>\BRIEF         Propogates the radiation fluxes through each level of the
!!               canopy.
!!
!! DESCRIPTION : This is an algorithm to follow radition fluxes through the
!!     canopy.  At each level, the path probabilities are determined by the
!!     raditional transfer scheme of Pinty et al (2006).  Notice that
!!     the fluxes given by this routine are all cumulative fluxes, not per
!!     level.
!!
!! RECENT CHANGE(S): None\n
!!
!! MAIN OUTPUT VARIABLE(S): ::Tran_Uncoll_Tot, ::Tran_Coll_Tot, ::Alb_Coll_Tot
!!
!! REFERENCE(S) :  
!!
!! FLOWCHART    : None
!_ ================================================================================================================================
SUBROUTINE propagate_fluxes(nlevels_loc, collim_down_cs, isotrop_down_cs, isotrop_up_cs, &
     Collim_Tran_UnColl_Unscaled, Collim_Tran_Coll_Unscaled, &
     Collim_Alb_Coll_Unscaled, Isotrop_Tran_UnColl_Unscaled, &
     Isotrop_Tran_Coll_Unscaled, Isotrop_Alb_Coll_Unscaled, br_base_temp_collim, &
     br_base_temp_isotrop, lisotrop, Tran_Uncoll_Tot, Tran_Coll_Tot, &
     Alb_Coll_Tot, lconverged, lprint)

  !! 0. Variables and parameter declaration
  !! 0.1 Input variables
  INTEGER,INTENT(IN)        :: nlevels_loc 
  REAL(r_std), DIMENSION(nlevels_loc),INTENT(IN)        :: Collim_Tran_UnColl_Unscaled
  REAL(r_std), DIMENSION(nlevels_loc),INTENT(IN)        :: Collim_Tran_Coll_Unscaled
  REAL(r_std), DIMENSION(nlevels_loc),INTENT(IN)        :: Collim_Alb_Coll_Unscaled
  REAL(r_std), DIMENSION(nlevels_loc),INTENT(IN)        :: Isotrop_Tran_UnColl_Unscaled
  REAL(r_std), DIMENSION(nlevels_loc),INTENT(IN)        :: Isotrop_Tran_Coll_Unscaled
  REAL(r_std), DIMENSION(nlevels_loc),INTENT(IN)        :: Isotrop_Alb_Coll_Unscaled
  REAL(r_std), INTENT(IN)    :: br_base_temp_collim           !! cosine of the solar zenith angle, between 0 and 1
  REAL(r_std), INTENT(IN)    :: br_base_temp_isotrop            !! cosine of the solar zenith angle, between 0 and 1
  LOGICAL, INTENT(IN)    :: lisotrop            !! are we dealing with an isotropic source?  only needed for correct partitioning
                                                !! of the collided and uncollided transmitted light...the total is not affected
  LOGICAL, INTENT(IN)    :: lprint              !! a flag to print

  !! 0.2 Output variables
  REAL(r_std), DIMENSION(nlevels_loc),INTENT(OUT)        :: Tran_Uncoll_Tot
  REAL(r_std), DIMENSION(nlevels_loc),INTENT(OUT)        :: Tran_Coll_Tot
  REAL(r_std), DIMENSION(nlevels_loc),INTENT(OUT)        :: Alb_Coll_Tot
  LOGICAL,INTENT(OUT)                                :: lconverged

  !! 0.3 Modified variables
  REAL(r_std), DIMENSION(0:nlevels_loc+1),INTENT(INOUT)    :: collim_down_cs
  REAL(r_std), DIMENSION(0:nlevels_loc+1),INTENT(INOUT)    :: isotrop_down_cs
  REAL(r_std), DIMENSION(0:nlevels_loc+1),INTENT(INOUT)    :: isotrop_up_cs

  !! 0.4 Local variables
  INTEGER :: istep,ilevel
  REAL(r_std),DIMENSION(0:nlevels_loc+1)            :: collim_down_ns,isotrop_down_ns,&
       isotrop_up_ns,Tran_Uncoll_Tot_temp
  
  !_ ================================================================================================================================

  Tran_Uncoll_Tot(:)=zero
  Tran_Coll_Tot(:)=zero
  Alb_Coll_Tot(:)=zero

  Tran_Uncoll_Tot_temp(:)=un

  istep=0
  DO

     istep=istep+1

     lconverged=.TRUE.

     ! Zero out the counters for the next step
     collim_down_ns(:)=zero
     isotrop_down_ns(:)=zero
     isotrop_up_ns(:)=zero


     ! Now we need to loop over all levels and see what light is entering the 
     ! level, and how it will propagate in the next step
     DO ilevel=1,nlevels_loc

        ! For collimated downwelling light into the level, it can be scattered 
        ! up, down, or pass through uncollided
        IF(collim_down_cs(ilevel) .GT. zero)THEN
           collim_down_ns(ilevel-1)=collim_down_ns(ilevel-1)+&
                collim_down_cs(ilevel)*Collim_Tran_UnColl_Unscaled(ilevel)

           ! This statement checks to see if this light has been previously
           ! scattered or not.  This term is only present in level nlevels_loc
           ! for the first step, level nlevels_loc-1 for the second step, 
           ! level nlevels_loc-2 for the third step, etc., and it only happens
           ! in the case of a collimated light source.
           IF(ilevel == nlevels_loc-istep+1 .AND. .NOT. lisotrop) THEN
              Tran_Uncoll_Tot_temp(ilevel)=Tran_Uncoll_Tot_temp(ilevel+1)*&
                   Collim_Tran_UnColl_Unscaled(ilevel)
           ENDIF

           isotrop_down_ns(ilevel-1)=isotrop_down_ns(ilevel-1)+&
                collim_down_cs(ilevel)*Collim_Tran_Coll_Unscaled(ilevel)
           isotrop_up_ns(ilevel+1)=isotrop_up_ns(ilevel+1)+&
                collim_down_cs(ilevel)*Collim_Alb_Coll_Unscaled(ilevel)
        ENDIF ! collim_down_cs(ilevel) .GT. zero

        ! For isotropic downwelling light, it can also be scattered up, down,
        ! or pass through uncollided
        IF(isotrop_down_cs(ilevel) .GT. zero)THEN
           isotrop_down_ns(ilevel-1)=isotrop_down_ns(ilevel-1)+&
                isotrop_down_cs(ilevel)*Isotrop_Tran_UnColl_Unscaled(ilevel)

           ! This is the same check as above, but this time the light has
           ! an isotropic source and not a collimated source
           IF(ilevel == nlevels_loc-istep+1 .AND. lisotrop) THEN
              Tran_Uncoll_Tot_temp(ilevel)=Tran_Uncoll_Tot_temp(ilevel+1)&
                   *Isotrop_Tran_UnColl_Unscaled(ilevel)
           ENDIF

           isotrop_down_ns(ilevel-1)=isotrop_down_ns(ilevel-1)+&
                isotrop_down_cs(ilevel)*Isotrop_Tran_Coll_Unscaled(ilevel)
           isotrop_up_ns(ilevel+1)=isotrop_up_ns(ilevel+1)+&
                isotrop_down_cs(ilevel)*Isotrop_Alb_Coll_Unscaled(ilevel)
        ENDIF ! isotrop_down_cs(ilevel) .GT. zero

        ! Isotropic upwelling light can pass through upwards collided or 
        ! uncollided with vegetation in this level, or it can be reflected downwards
        IF(isotrop_up_cs(ilevel) .GT. zero)THEN
           isotrop_up_ns(ilevel+1)=isotrop_up_ns(ilevel+1)+isotrop_up_cs(ilevel)&
                *Isotrop_Tran_UnColl_Unscaled(ilevel)
           isotrop_up_ns(ilevel+1)=isotrop_up_ns(ilevel+1)+isotrop_up_cs(ilevel)&
                *Isotrop_Tran_Coll_Unscaled(ilevel)
           isotrop_down_ns(ilevel-1)=isotrop_down_ns(ilevel-1)+&
                isotrop_up_cs(ilevel)*Isotrop_Alb_Coll_Unscaled(ilevel)
        ENDIF

        ! there can be no collimated upwards light, since upwards light must 
        ! always have reflected off something


     ENDDO ! ilevel=1,nlevels_loc

     ! The background is a bit special. there is no transmission, but there is 
     ! reflection, which leads to a source term to the bottom level from below.
     isotrop_up_ns(1)=isotrop_up_ns(1)+collim_down_cs(0)*br_base_temp_collim
     isotrop_up_ns(1)=isotrop_up_ns(1)+isotrop_down_cs(0)*br_base_temp_isotrop


     ! Now we add the light generated here to our cumulative counters to track
     ! the total amount that leaves the canopy, either through being absorbed
     ! by the background or being reflected back into the atmosphere.
     ! We keep track of the uncoll above, so here we track the total light that
     ! is transmitted to the soil, and then at the end we taken the difference 
     ! to get the collided radation.
     Tran_Coll_Tot(1:nlevels_loc)=Tran_Coll_Tot(1:nlevels_loc)+&
          isotrop_down_ns(0:nlevels_loc-1)+collim_down_ns(0:nlevels_loc-1)
     Alb_Coll_Tot(1:nlevels_loc)=Alb_Coll_Tot(1:nlevels_loc)+isotrop_up_ns(2:nlevels_loc+1)

     ! now we update the light we are currently tracking
     collim_down_cs(:)=collim_down_ns(:)
     isotrop_down_cs(:)=isotrop_down_ns(:)
     isotrop_up_cs(:)=isotrop_up_ns(:)




     ! check for convegence...if all the values of light are currently below a 
     ! threshold, we're probably good assuming the threshold is low enough.
     IF(MAXVAL(collim_down_cs,1) .GT. alb_threshold) lconverged=.FALSE.
     IF(MAXVAL(isotrop_down_cs,1) .GT. alb_threshold) lconverged=.FALSE.
     IF(MAXVAL(isotrop_up_cs,1) .GT. alb_threshold) lconverged=.FALSE.

     IF(lconverged)THEN
        IF(lprint) WRITE(numout,*) 'Converged after this many steps: ',istep
        EXIT
     ENDIF

     ! This number here could also be externalized.
     IF(istep .GE. 1000)THEN
        WRITE(numout,*) '*********************************************************'
        WRITE(numout,'(A,I6,F14.10)') 'Albedo not converging!',istep,alb_threshold
        WRITE(numout,'(A)') '                  collim_down_cs ' // &
             'isotrop_down_cs  isotrop_up_cs'
        DO ilevel=0,nlevels_loc+1
           WRITE(numout,'(I4,3F14.10)') ilevel, &
                collim_down_cs(ilevel),isotrop_down_cs(ilevel),isotrop_up_cs(ilevel)
        ENDDO
        WRITE(numout,*) 'You should increase either the number' // &
             'of steps or the alb_threshold.'
        WRITE(numout,*) '*********************************************************'
        EXIT
     ENDIF ! istep .GE. 1000
  ENDDO ! convergence loop


  ! now separate the collided from the uncollided light.  Notice that
  ! this is not really needed for any purposes other than debugging, as the
  ! important quantity is the total amount of light striking the ground.
  Tran_UnColl_Tot(1:nlevels_loc)=Tran_UnColl_Tot_temp(1:nlevels_loc)
  Tran_Coll_Tot(1:nlevels_loc)=Tran_Coll_Tot(1:nlevels_loc)-Tran_UnColl_Tot(1:nlevels_loc)

  ! Some debugging information
  IF(lprint)THEN
     WRITE(numout,'(A)') '      Tran_Uncoll_Tot   Tran_Coll_Tot  Alb_Coll_Tot'
     DO ilevel=nlevels_loc,1,-1
        WRITE(numout,'(I4,3F10.6)') ilevel, &
             Tran_Uncoll_Tot(ilevel),Tran_Coll_Tot(ilevel),Alb_Coll_Tot(ilevel)
     ENDDO
  ENDIF

END SUBROUTINE propagate_fluxes


END MODULE albedo
